"""

remap_detection_categories.py

Given a MegaDetector results file, remap the category IDs according to a specified
dictionary, writing the results to a new file.

Currently only supports remapping detection categories, not classification categories.

"""

#%% Constants and imports

import os
import json

from tqdm import tqdm

from megadetector.utils.ct_utils import invert_dictionary
from megadetector.utils.ct_utils import write_json


#%% Main function

def remap_detection_categories(input_file,
                               output_file,
                               target_category_map,
                               input_category_name_to_output_category_name,
                               overwrite=False,
                               invalid_category_handling='unknown'):
    """
    Given a MegaDetector results file [input_file], remap the category IDs according to the dictionary
    [target_category_map], writing the results to [output_file].  The remapped dictionary needs to have
    the same category names as the input file's detection_categories dictionary.

    Typically used to map, e.g., a variety of species to the class "mammal" or the class "animal".

    Currently only supports remapping detection categories, not classification categories.

    Args:
        input_file (str): the MD .json results file to remap
        output_file (str): the remapped .json file to write
        target_category_map (dict): the category mapping that should be used in the output file,
            mapping string-ints to class names. This can also be a MD results file, in which case
            we'll use that file's detection_categories dictionary.
        input_category_name_to_output_category_name (dict): str->str, the specific category mapping
            that should be used, otherwise will determine from target class names
        overwrite (bool, optional): whether to overwrite [output_file] if it exists; if this is True and
            [output_file] exists, this function is a no-op
        invalid_category_handling (str, optional): what to do about categories that are not in
            the input file's category list ('error' or 'unknown'), if 'unknown', creates a new
            "unknown" category

    """

    if os.path.exists(output_file) and (not overwrite):
        print('File {} exists, bypassing remapping'.format(output_file))
        return

    assert invalid_category_handling in ('error','unknown'), \
        'Invalid value for invalid_category_handling: {}'.format(invalid_category_handling)

    assert os.path.isfile(input_file), \
        'File {} does not exist'.format(input_file)

    # If "target_category_map" is passed as a filename, load the "detection_categories"
    # dict.
    if isinstance(target_category_map,str):
        target_categories_file = target_category_map
        with open(target_categories_file,'r') as f:
            d = json.load(f)
            target_category_map = d['detection_categories']
    assert isinstance(target_category_map,dict)

    with open(input_file,'r') as f:
        input_data = json.load(f)

    input_images = input_data['images']
    input_categories = input_data['detection_categories']

    output_category_name_to_output_category_id = invert_dictionary(target_category_map)

    # Map input category IDs to output category IDs
    input_category_id_to_output_category_id = {}
    for input_category_id in input_categories.keys():
        input_category_name = input_categories[input_category_id]
        output_category_name = \
            input_category_name_to_output_category_name[input_category_name]
        output_category_id = \
            output_category_name_to_output_category_id[output_category_name]
        input_category_id_to_output_category_id[input_category_id] = output_category_id

    # im = input_images[0]
    for im in tqdm(input_images):

        if 'detections' not in im or im['detections'] is None:
            continue

        # det = im['detections'][0]
        for det in im['detections']:
            input_category_id = det['category']
            if input_category_id not in input_category_id_to_output_category_id:
                input_category_name = '[unknown]'
                if input_category_id in input_categories:
                    input_category_name = input_categories[input_category_id]
                s = 'Detection category {} ({}) not mapped'.format(
                    input_category_id,input_category_name)
                if invalid_category_handling == 'error':
                    raise ValueError(s)
                else:
                    assert invalid_category_handling == 'unknown'
                    print('Warning: {}'.format(s))
                    if 'unknown' in output_category_name_to_output_category_id:
                        output_category_id = output_category_name_to_output_category_id['unknown']
                    else:
                        category_id_values = [int(x) for x in output_category_name_to_output_category_id.values()]
                        unknown_category_id = max(category_id_values) + 1
                        output_category_name_to_output_category_id['unknown'] = unknown_category_id
                        output_category_id = unknown_category_id
            else:
                output_category_id = input_category_id_to_output_category_id[input_category_id]
            det['category'] = output_category_id

    # ...for each image

    input_data['detection_categories'] = target_category_map

    write_json(output_file,input_data)

    print('Saved remapped results to {}'.format(output_file))

# ...def remap_detection_categories(...)



"""

run_md_and_speciesnet.py

Script to run MegaDetector and SpeciesNet on a folder of images and/or videos.
Runs MD first, then runs SpeciesNet on every above-threshold crop.

"""

#%% Constants, imports, environment

import argparse
import json
import multiprocessing
import os
import sys
import time

from tqdm import tqdm
from multiprocessing import JoinableQueue, Process, Queue
from threading import Thread

import humanfriendly

from megadetector.detection import run_detector_batch
from megadetector.detection.video_utils import find_videos, run_callback_on_frames, is_video_file
from megadetector.detection.run_detector_batch import load_and_run_detector_batch
from megadetector.detection.run_detector import DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD
from megadetector.detection.run_detector import CONF_DIGITS
from megadetector.detection.run_detector_batch import write_results_to_file
from megadetector.utils.ct_utils import round_float
from megadetector.utils.ct_utils import write_json
from megadetector.utils.ct_utils import make_temp_folder
from megadetector.utils.ct_utils import is_list_sorted
from megadetector.utils.ct_utils import is_sphinx_build
from megadetector.utils.ct_utils import args_to_object
from megadetector.utils.path_utils import find_images
from megadetector.utils.path_utils import test_file_write
from megadetector.visualization import visualization_utils as vis_utils
from megadetector.postprocessing.validate_batch_results import \
    validate_batch_results, ValidateBatchResultsOptions
from megadetector.detection.process_video import \
    process_videos, ProcessVideoOptions
from megadetector.postprocessing.combine_batch_outputs import combine_batch_output_files

# We aren't taking an explicit dependency on the speciesnet package yet,
# so we wrap this in a try/except so sphinx can still document this module.
try:
    from speciesnet import SpeciesNetClassifier
    from speciesnet.utils import BBox
    from speciesnet.ensemble import SpeciesNetEnsemble
    from speciesnet.geofence_utils import roll_up_labels_to_first_matching_level
    from speciesnet.geofence_utils import geofence_animal_classification
except Exception:
    pass


#%% Constants

DEFAULT_DETECTOR_MODEL = 'MDV5A'
try:
    from speciesnet import DEFAULT_MODEL as DEFAULT_CLASSIFIER_MODEL
except Exception:
    DEFAULT_CLASSIFIER_MODEL = 'kaggle:google/speciesnet/pyTorch/v4.0.2a'
DEFAULT_DETECTION_CONFIDENCE_THRESHOLD_FOR_CLASSIFICATION = 0.1
DEFAULT_DETECTION_CONFIDENCE_THRESHOLD_FOR_OUTPUT = DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD
DEFAULT_DETECTOR_BATCH_SIZE = 1
DEFAULT_CLASSIFIER_BATCH_SIZE = 8
DEFAULT_LOADER_WORKERS = 4
DEFAULT_WORKER_TYPE = 'thread'

# This determines the maximum number of image filenames that can be assigned to
# each of the producer workers before blocking.  The actual size of the queue
# will be MAX_IMAGE_QUEUE_SIZE_PER_WORKER * n_workers.  This is only used for
# the classification step.
MAX_IMAGE_QUEUE_SIZE_PER_WORKER = 30

# This determines the maximum number of crops that can accumulate in the queue
# used to communicate between the producers (which read and crop images) and the
# consumer (which runs the classifier).  This is only used for the classification step.
MAX_BATCH_QUEUE_SIZE = 300

# Default interval between frames we should process when processing video.
# This is only used for the detection step.
DEFAULT_SECONDS_PER_VIDEO_FRAME = 1.0

# Max number of classification scores to include per detection
DEFAULT_TOP_N_SCORES = 2

# Unless --norollup is specified, roll up taxonomic levels until the
# cumulative confidence is above this value.  Only relevant when
# geofencing is disabled, otherwise the default speciesnet library
# constants are used.
DEFAULT_ROLLUP_TARGET_CONFIDENCE = 0.65

# When the called supplies an existing MD results file, should we validate it before
# starting classification?
VALIDATE_DETECTION_FILE = False

verbose = False


#%% Main options class

class RunMDSpeciesNetOptions:
    """
    Class controlling the behavior of run_md_and_speciesnet()
    """

    def __init__(self):

        #: Folder containing images and/or videos to process
        self.source = None

        #: Output file for results (JSON format)
        self.output_file = None

        #: What to do if the output file exists ('overwrite', 'error', 'skip')
        self.overwrite_handling = 'overwrite'

        #: MegaDetector model identifier (MDv5a, MDv5b, MDv1000-redwood, etc.)
        self.detector_model = DEFAULT_DETECTOR_MODEL

        #: SpeciesNet classifier model identifier (e.g. kaggle:google/speciesnet/pyTorch/v4.0.2a)
        self.classification_model = DEFAULT_CLASSIFIER_MODEL

        #: Batch size for MegaDetector inference
        self.detector_batch_size = DEFAULT_DETECTOR_BATCH_SIZE

        #: Batch size for SpeciesNet classification
        self.classifier_batch_size = DEFAULT_CLASSIFIER_BATCH_SIZE

        #: Number of worker threads for preprocessing
        self.loader_workers = DEFAULT_LOADER_WORKERS

        #: Classify detections above this threshold
        self.detection_confidence_threshold_for_classification = \
            DEFAULT_DETECTION_CONFIDENCE_THRESHOLD_FOR_CLASSIFICATION

        #: Include detections above this threshold in the output
        self.detection_confidence_threshold_for_output = \
            DEFAULT_DETECTION_CONFIDENCE_THRESHOLD_FOR_OUTPUT

        #: Folder for intermediate files (default: system temp)
        self.intermediate_file_folder = None

        #: Keep intermediate files (e.g. detection-only results file)
        self.keep_intermediate_files = False

        #: Disable taxonomic rollup
        self.norollup = False

        #: Target confidence threshold for taxonomic rollup
        self.rollup_target_confidence = DEFAULT_ROLLUP_TARGET_CONFIDENCE

        #: Country code (ISO 3166-1 alpha-3) for geofencing (default None, no geoferencing)
        self.country = None

        #: Admin1 region/state code for geofencing
        self.admin1_region = None

        #: Path to existing MegaDetector output file (skips detection step)
        self.detections_file = None

        #: Ignore videos, only process images
        self.skip_video = False

        #: Ignore images, only process videos
        self.skip_images = False

        #: Sample every Nth frame from videos
        #:
        #: Mutually exclusive with time_sample
        self.frame_sample = None

        #: Sample frames every N seconds from videos
        #:
        #: Mutually exclusive with frame_sample
        self.time_sample = DEFAULT_SECONDS_PER_VIDEO_FRAME

        #: Enable additional debug output
        self.verbose = False

        #: Worker type for parallelization; should be "thread" or "process"
        self.worker_type = DEFAULT_WORKER_TYPE

        #: Include raw (pre-rollup/geofence) classification scores in output
        self.include_raw_classifications = False

        if self.time_sample is None and self.frame_sample is None:
            self.time_sample = DEFAULT_SECONDS_PER_VIDEO_FRAME

# ...class RunMDSpeciesNetOptions


#%% Support classes

class CropMetadata:
    """
    Metadata for a crop extracted from an image detection.
    """

    def __init__(self,
                 image_file: str,
                 detection_index: int,
                 bbox: list[float],
                 original_width: int,
                 original_height: int):
        """
        Args:
            image_file (str): path to the original image file
            detection_index (int): index of this detection in the image
            bbox (List[float]): normalized bounding box [x_min, y_min, width, height]
            original_width (int): width of the original image
            original_height (int): height of the original image
        """

        self.image_file = image_file
        self.detection_index = detection_index
        self.bbox = bbox
        self.original_width = original_width
        self.original_height = original_height


class CropBatch:
    """
    A batch of crops with their metadata for classification.
    """

    def __init__(self):
        #: List of preprocessed images
        self.crops = []

        #: List of CropMetadata objects
        self.metadata = []

    def add_crop(self, crop_data, metadata):
        """
        Args:
            crop_data (PreprocessedImage): preprocessed image data from
                SpeciesNetClassifier.preprocess()
            metadata (CropMetadata): metadata for this crop
        """

        self.crops.append(crop_data)
        self.metadata.append(metadata)

    def __len__(self):
        return len(self.crops)


#%% Support functions for classification

def _process_image_detections(file_path: str,
                              absolute_file_path: str,
                              detection_results: dict,
                              classifier: 'SpeciesNetClassifier',
                              detection_confidence_threshold: float,
                              batch_queue: Queue):
    """
    Process detections from a single image.

    Args:
        file_path (str): relative path to the image file
        absolute_file_path (str): absolute path to the image file
        detection_results (dict): detection results for this image
        classifier (SpeciesNetClassifier): classifier instance for preprocessing
        detection_confidence_threshold (float): classify detections above this threshold
        batch_queue (Queue): queue to send crops to
    """

    detections = detection_results['detections']

    # Don't bother loading images that have no above-threshold detections
    detections_above_threshold = \
        [d for d in detections if d['conf'] >= detection_confidence_threshold]
    if len(detections_above_threshold) == 0:
        return

    # Load the image
    try:
        image = vis_utils.load_image(absolute_file_path)
        original_width, original_height = image.size
    except Exception as e:
        print('Warning: failed to load image {}: {}'.format(file_path, str(e)))

        # Send failure information to consumer
        failure_metadata = CropMetadata(
            image_file=file_path,
            detection_index=-1,  # -1 indicates whole-image failure
            bbox=[],
            original_width=0,
            original_height=0
        )
        batch_queue.put(('failure',
                         'Failed to load image: {}'.format(str(e)),
                         failure_metadata))
        return

    # Process each detection above threshold
    #
    # detection_index needs to index into the original list of detections
    # (this is how classification results will be associated with detections
    # later), so iterate over "detections" here, rather than
    # "detections_above_threshold".
    for detection_index, detection in enumerate(detections):

        conf = detection['conf']
        if conf < detection_confidence_threshold:
            continue

        bbox = detection['bbox']
        assert len(bbox) == 4

        # Convert normalized bbox to BBox object for SpeciesNet
        speciesnet_bbox = BBox(
            xmin=bbox[0],
            ymin=bbox[1],
            width=bbox[2],
            height=bbox[3]
        )

        # Preprocess the crop
        try:

            preprocessed_crop = classifier.preprocess(
                image,
                bboxes=[speciesnet_bbox],
                resize=True
            )

            if preprocessed_crop is not None:

                metadata = CropMetadata(
                    image_file=file_path,
                    detection_index=detection_index,
                    bbox=bbox,
                    original_width=original_width,
                    original_height=original_height
                )

                # Send individual crop to the consumer
                batch_queue.put(('crop', preprocessed_crop, metadata))

        except Exception as e:

            print('Warning: failed to preprocess crop from {}, detection {}: {}'.format(
                file_path, detection_index, str(e)))

            # Send failure information to consumer
            failure_metadata = CropMetadata(
                image_file=file_path,
                detection_index=detection_index,
                bbox=bbox,
                original_width=original_width,
                original_height=original_height
            )
            batch_queue.put(('failure',
                             'Failed to preprocess crop: {}'.format(str(e)),
                             failure_metadata))

        # ...try/except

    # ...for each detection in this image

# ...def _process_image_detections(...)


def _process_video_detections(file_path: str,
                              absolute_file_path: str,
                              detection_results: dict,
                              classifier: 'SpeciesNetClassifier',
                              detection_confidence_threshold: float,
                              batch_queue: Queue):
    """
    Process detections from a single video.

    Args:
        file_path (str): relative path to the video file
        absolute_file_path (str): absolute path to the video file
        detection_results (dict): detection results for this video
        classifier (SpeciesNetClassifier): classifier instance for preprocessing
        detection_confidence_threshold (float): classify detections above this threshold
        batch_queue (Queue): queue to send crops to
    """

    detections = detection_results['detections']

    # Find frames with above-threshold detections
    frames_with_detections = set()
    frame_to_detections = {}

    for detection_index, detection in enumerate(detections):

        conf = detection['conf']
        if conf < detection_confidence_threshold:
            continue

        frame_number = detection['frame_number']
        frames_with_detections.add(frame_number)

        if frame_number not in frame_to_detections:
            frame_to_detections[frame_number] = []
        frame_to_detections[frame_number].append((detection_index, detection))

    # ...for each detection in this video

    if len(frames_with_detections) == 0:
        return

    frames_to_process = sorted(list(frames_with_detections))

    # Define callback for processing each frame
    def frame_callback(frame_array, frame_id):
        """
        Callback to process a single frame.

        Args:
            frame_array (numpy.ndarray): frame data in PIL format
            frame_id (str): frame identifier like "frame0006.jpg"
        """

        # Extract frame number from frame_id (e.g., "frame0006.jpg" -> 6)
        import re
        match = re.match(r'frame(\d+)\.jpg', frame_id)
        if not match:
            print('Warning: could not parse frame number from {}'.format(frame_id))
            return
        frame_number = int(match.group(1))

        # Only process frames for which we have detection results
        if frame_number not in frame_to_detections:
            return

        # Convert numpy array to PIL Image
        from PIL import Image
        if frame_array.dtype != 'uint8':
            frame_array = (frame_array * 255).astype('uint8')
        frame_image = Image.fromarray(frame_array)
        original_width, original_height = frame_image.size

        # Process each detection in this frame
        for detection_index, detection in frame_to_detections[frame_number]:

            bbox = detection['bbox']
            assert len(bbox) == 4

            # Convert normalized bbox to BBox object for SpeciesNet
            speciesnet_bbox = BBox(
                xmin=bbox[0],
                ymin=bbox[1],
                width=bbox[2],
                height=bbox[3]
            )

            # Preprocess the crop
            try:

                preprocessed_crop = classifier.preprocess(
                    frame_image,
                    bboxes=[speciesnet_bbox],
                    resize=True
                )

                if preprocessed_crop is not None:
                    metadata = CropMetadata(
                        image_file=file_path,
                        detection_index=detection_index,
                        bbox=bbox,
                        original_width=original_width,
                        original_height=original_height
                    )

                    # Send individual crop immediately to consumer
                    batch_queue.put(('crop', preprocessed_crop, metadata))

            except Exception as e:

                print('Warning: failed to preprocess crop from {}, detection {}: {}'.format(
                    file_path, detection_index, str(e)))

                # Send failure information to consumer
                failure_metadata = CropMetadata(
                    image_file=file_path,
                    detection_index=detection_index,
                    bbox=bbox,
                    original_width=original_width,
                    original_height=original_height
                )
                batch_queue.put(('failure',
                                 'Failed to preprocess crop: {}'.format(str(e)),
                                 failure_metadata))

            # ...try/except

        # ...for each detection

    # ...def frame_callback(...)

    # Process the video frames
    try:

        run_callback_on_frames(
            input_video_file=absolute_file_path,
            frame_callback=frame_callback,
            frames_to_process=frames_to_process,
            verbose=verbose
        )

    except Exception as e:

        print('Warning: failed to process video {}: {}'.format(file_path, str(e)))

        # Send failure information to consumer for the whole video
        failure_metadata = CropMetadata(
            image_file=file_path,
            detection_index=-1,  # -1 indicates whole-file failure
            bbox=[],
            original_width=0,
            original_height=0
        )
        batch_queue.put(('failure',
                         'Failed to process video: {}'.format(str(e)),
                         failure_metadata))
    # ...try/except

# ...def _process_video_detections(...)


def _crop_producer_func(image_queue: JoinableQueue,
                        batch_queue: Queue,
                        classifier_model: str,
                        detection_confidence_threshold: float,
                        source_folder: str,
                        producer_id: int = -1,
                        preloaded_classifier: 'SpeciesNetClassifier' = None):
    """
    Producer function for classification workers.

    Reads images and videos from [image_queue], crops detections above a threshold,
    preprocesses them, and sends individual crops to [batch_queue].
    See the documentation of _crop_consumer_func to for the format of the
    tuples placed on batch_queue.

    Args:
        image_queue (JoinableQueue): queue containing detection_results dicts (for both images and videos)
        batch_queue (Queue): queue to put individual crops into
        classifier_model (str): classifier model identifier to load in this process
        detection_confidence_threshold (float): classify detections above this threshold
        source_folder (str): source folder to resolve relative paths
        producer_id (int, optional): identifier for this producer worker
        preloaded_classifier (SpeciesNetClassifier, optional): pre-loaded classifier instance
            (for thread-based workers, to avoid loading models in threads)
    """

    if verbose:
        print('Classification producer starting: ID {}'.format(producer_id))

    # Load classifier; this is just being used as a preprocessor, so we force device=cpu.
    #
    # When using threads, we pre-load the classifier in the main thread to avoid PyTorch FX
    # issues with loading models in worker threads.
    if preloaded_classifier is not None:
        classifier = preloaded_classifier
        if verbose:
            print('Classification producer {}: using pre-loaded classifier'.format(producer_id))
    else:
        # There are a number of reasons loading the model might fail; note to self: *don't*
        # catch Exceptions here.  This should be a catastrophic failure that stops the whole
        # process.
        classifier = SpeciesNetClassifier(classifier_model, device='cpu')
        if verbose:
            print('Classification producer {}: loaded classifier'.format(producer_id))

    while True:

        # Pull an image of detection results from the queue
        detection_results = image_queue.get()

        # Pulling None from the queue indicates that this producer is done
        if detection_results is None:
            image_queue.task_done()
            break

        file_path = detection_results['file']

        # Skip files that failed at the detection stage
        if 'failure' in detection_results:
            image_queue.task_done()
            continue

        # Skip files with no detections
        detections = detection_results['detections']
        if len(detections) == 0:
            image_queue.task_done()
            continue

        # Determine if this is an image or video
        absolute_file_path = os.path.join(source_folder, file_path)
        is_video = is_video_file(file_path)

        if is_video:

            # Process video
            _process_video_detections(
                file_path=file_path,
                absolute_file_path=absolute_file_path,
                detection_results=detection_results,
                classifier=classifier,
                detection_confidence_threshold=detection_confidence_threshold,
                batch_queue=batch_queue
            )

        else:

            # Process image
            _process_image_detections(
                file_path=file_path,
                absolute_file_path=absolute_file_path,
                detection_results=detection_results,
                classifier=classifier,
                detection_confidence_threshold=detection_confidence_threshold,
                batch_queue=batch_queue
            )

        image_queue.task_done()

    # ...while(we still have items to process)

    # Send sentinel to indicate this producer is done
    batch_queue.put(None)

    if verbose:
        print('Classification producer {} finished'.format(producer_id))

# ...def _crop_producer_func(...)


def _crop_consumer_func(batch_queue: Queue,
                        results_queue: Queue,
                        classifier_model: str,
                        batch_size: int,
                        num_producers: int,
                        enable_rollup: bool,
                        country: str = None,
                        admin1_region: str = None,
                        preloaded_classifier: 'SpeciesNetClassifier' = None,
                        rollup_target_confidence: float = DEFAULT_ROLLUP_TARGET_CONFIDENCE):
    """
    Consumer function for classification inference.

    Pulls individual crops from batch_queue, assembles them into batches,
    runs inference, and puts results into results_queue.

    Args:
        batch_queue (Queue): queue containing individual crop tuples or failures.
            Items on this queue are either None (to indicate that a producer finished)
            or tuples formatted as (type,image,metadata).  [type] is a string (either
            "crop" or "failure"), [image] is a PreprocessedImage, and [metadata] is
            a CropMetadata object.
        results_queue (Queue): queue to put classification results into
        classifier_model (str): classifier model identifier to load
        batch_size (int): batch size for inference
        num_producers (int): number of producer workers
        enable_rollup (bool): whether to apply taxonomic rollup
        country (str, optional): country code for geofencing
        admin1_region (str, optional): admin1 region for geofencing
        preloaded_classifier (SpeciesNetClassifier, optional): pre-loaded classifier instance
            (for thread-based workers, to avoid loading models in threads)
        rollup_target_confidence (float, optional): target confidence threshold for taxonomic
            rollup.  Ignored if enable_rollup is False.
    """

    if verbose:
        print('Classification consumer starting')

    # Load classifier
    # When using threads, we pre-load the classifier in the main thread to avoid PyTorch FX
    # issues with loading models in worker threads.
    if preloaded_classifier is not None:
        classifier = preloaded_classifier
        if verbose:
            print('Classification consumer: using pre-loaded classifier')
    else:
        try:
            classifier = SpeciesNetClassifier(classifier_model)
            if verbose:
                print('Classification consumer: loaded classifier')
        except Exception as e:
            print('Classification consumer: failed to load classifier: {}'.format(str(e)))
            results_queue.put({})
            return

    all_results = {}  # image_file -> {detection_index -> classification_result}
    current_batch = CropBatch()
    producers_finished = 0

    # Load ensemble metadata if rollup/geofencing is enabled
    taxonomy_map = {}
    geofence_map = {}

    if (enable_rollup) or (country is not None):

        # Note to self: there are a number of reasons loading the ensemble
        # could fail here; don't catch this exception, this should be a
        # catatstrophic failure.
        ensemble = SpeciesNetEnsemble(
            classifier_model, geofence=(country is not None))
        taxonomy_map = ensemble.taxonomy_map
        geofence_map = ensemble.geofence_map

    # ...if we need to load ensemble components

    while True:

        # Pull an item from the queue
        item = batch_queue.get()

        # This indicates that a producer worker finished
        if item is None:

            producers_finished += 1
            if producers_finished == num_producers:
                # Process any remaining images
                if len(current_batch) > 0:
                    _process_classification_batch(
                        current_batch, classifier, all_results,
                        enable_rollup, taxonomy_map, geofence_map,
                        country, admin1_region, rollup_target_confidence
                    )
                break
            continue

        # ...if a producer finished

        # If we got here, we know we have a crop to process, or
        # a failure to ignore.
        assert isinstance(item, tuple) and len(item) == 3
        item_type, data, metadata = item

        if metadata.image_file not in all_results:
            all_results[metadata.image_file] = {}

        # We should never be processing the same detection twice
        assert metadata.detection_index not in all_results[metadata.image_file]

        if item_type == 'failure':

            all_results[metadata.image_file][metadata.detection_index] = {
                'failure': 'Failure classification: {}'.format(data)
            }

        else:

            assert item_type == 'crop'
            current_batch.add_crop(data, metadata)
            assert len(current_batch) <= batch_size

            # Process batch if necessary
            if len(current_batch) == batch_size:
                _process_classification_batch(
                    current_batch, classifier, all_results,
                    enable_rollup, taxonomy_map, geofence_map,
                    country, admin1_region, rollup_target_confidence
                )
                current_batch = CropBatch()

        # ...was this item a failure or a crop?

    # ...while (we have items to process)

    # Send all the results at once back to the main process
    results_queue.put(all_results)

    if verbose:
        print('Classification consumer finished')

# ...def _crop_consumer_func(...)


def _process_classification_batch(batch: CropBatch,
                                  classifier: 'SpeciesNetClassifier',
                                  all_results: dict,
                                  enable_rollup: bool,
                                  taxonomy_map: dict,
                                  geofence_map: dict,
                                  country: str = None,
                                  admin1_region: str = None,
                                  rollup_target_confidence: float =
                                    DEFAULT_ROLLUP_TARGET_CONFIDENCE):
    """
    Run a batch of crops through the classifier.

    Args:
        batch (CropBatch): batch of crops to process
        classifier (SpeciesNetClassifier): classifier instance
        all_results (dict): dictionary to store results in, modified in-place with format:
            {image_file: {detection_index: {'predictions': [[class_name, score], ...]}
            or {image_file: {detection_index: {'failure': error_message}}}.
        enable_rollup (bool): whether to apply rollup
        taxonomy_map (dict): taxonomy mapping for rollup
        geofence_map (dict): geofence mapping
        country (str, optional): country code for geofencing
        admin1_region (str, optional): admin1 region for geofencing
        rollup_target_confidence (float, optional): target confidence threshold for
            taxonomic rollup, ignored if enable_rollup is False
    """

    if len(batch) == 0:
        print('Warning: _process_classification_batch received empty batch')
        return

    # Prepare batch for inference
    filepaths = [f"{metadata.image_file}_{metadata.detection_index}"
                 for metadata in batch.metadata]

    # Run batch inference
    try:
        batch_results = classifier.batch_predict(filepaths, batch.crops)
    except Exception as e:
        print('*** Batch classification failed: {} ***'.format(str(e)))
        # Mark all crops in this batch as failed
        for metadata in batch.metadata:
            if metadata.image_file not in all_results:
                all_results[metadata.image_file] = {}
            all_results[metadata.image_file][metadata.detection_index] = {
                'failure': 'Failure classification: {}'.format(str(e))
            }
        return

    # Process results
    assert len(batch_results) == len(batch.metadata)
    assert len(batch_results) == len(filepaths)

    for i_result in range(0, len(batch_results)):

        filepath = filepaths[i_result]
        result = batch_results[i_result]
        metadata = batch.metadata[i_result]

        assert metadata.image_file in all_results, \
            'File {} not in results dict'.format(metadata.image_file)

        detection_index = metadata.detection_index

        # Handle classification failure
        if 'failures' in result:
            print('*** Classification failure for image: {} ***'.format(
                filepath))
            all_results[metadata.image_file][detection_index] = {
                'failure': 'Failure classification: SpeciesNet classifier failed'
            }
            continue

        # Extract classification results; this is a dict with keys "classes"
        # and "scores", each of which points to a list.
        classifications = result['classifications']
        classes = classifications['classes']
        scores = classifications['scores']

        classification_was_geofenced = False

        predicted_class = classes[0]
        predicted_score = scores[0]

        # Possibly apply geofencing
        if country:

            geofence_result = geofence_animal_classification(
                labels=classes,
                scores=scores,
                country=country,
                admin1_region=admin1_region,
                taxonomy_map=taxonomy_map,
                geofence_map=geofence_map,
                enable_geofence=True
            )

            geofenced_class, geofenced_score, prediction_source = geofence_result

            if prediction_source != 'classifier':
                classification_was_geofenced = True
                predicted_class = geofenced_class
                predicted_score = geofenced_score

        # ...if we might need to apply geofencing

        # Possibly apply rollup; this was already done if geofencing was applied
        if enable_rollup and (not classification_was_geofenced):

            rollup_result = roll_up_labels_to_first_matching_level(
                labels=classes,
                scores=scores,
                country=country,
                admin1_region=admin1_region,
                target_taxonomy_levels=['species','genus','family','order','class','kingdom'],
                non_blank_threshold=rollup_target_confidence,
                taxonomy_map=taxonomy_map,
                geofence_map=geofence_map,
                enable_geofence=(country is not None)
            )

            if rollup_result is not None:
                rolled_up_class, rolled_up_score, prediction_source = rollup_result
                predicted_class = rolled_up_class
                predicted_score = rolled_up_score

        # ...if we might need to apply taxonomic rollup

        # For now, we'll store category names as strings; these will be assigned to integer
        # IDs before writing results to file later.
        classification = [predicted_class,predicted_score]

        # Also report raw model classifications
        raw_classifications = []
        for i_class in range(0,len(classes)):
            raw_classifications.append([classes[i_class],scores[i_class]])

        all_results[metadata.image_file][detection_index] = {
            'classifications': [classification],
            'raw_classifications': raw_classifications
        }

    # ...for each result in this batch

# ...def _process_classification_batch(...)


#%% Inference functions

def _run_detection_step(source_folder: str,
                        detector_output_file: str,
                        detector_model: str = DEFAULT_DETECTOR_MODEL,
                        detector_batch_size: int = DEFAULT_DETECTOR_BATCH_SIZE,
                        detection_confidence_threshold: float = DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD,
                        detector_worker_threads: int = DEFAULT_LOADER_WORKERS,
                        worker_type: str = DEFAULT_WORKER_TYPE,
                        skip_images: bool = False,
                        skip_video: bool = False,
                        frame_sample: int = None,
                        time_sample: float = None) -> str:
    """
    Run MegaDetector on all images/videos in [source_folder].

    Args:
        source_folder (str): folder containing images/videos
        detector_output_file (str): output .json file
        detector_model (str, optional): detector model identifier
        detector_batch_size (int, optional): batch size for detection
        detection_confidence_threshold (float, optional): confidence threshold for detections
            (to include in the output file)
        detector_worker_threads (int, optional): number of workers to use for preprocessing
        worker_type (str, optional): type of worker parallelization ("thread" or "process")
        skip_images (bool, optional): ignore images, only process videos
        skip_video (bool, optional): ignore videos, only process images
        frame_sample (int, optional): sample every Nth frame from videos
        time_sample (float, optional): sample frames every N seconds from videos
    """

    print('Starting detection step...')

    # Validate arguments
    assert not (frame_sample is None and time_sample is None), \
        'Must specify either frame_sample or time_sample'

    # Find image and video files
    if not skip_images:
        image_files = find_images(source_folder, recursive=True,
                                  return_relative_paths=False)
    else:
        image_files = []

    if not skip_video:
        video_files = find_videos(source_folder, recursive=True,
                                  return_relative_paths=False)
    else:
        video_files = []

    if len(image_files) == 0 and len(video_files) == 0:
        raise ValueError(
            'No images or videos found in {}'.format(source_folder))

    print('Found {} images and {} videos'.format(len(image_files), len(video_files)))

    files_to_merge = []

    # Process images if necessary
    if len(image_files) > 0:

        print('Running MegaDetector on {} images...'.format(len(image_files)))

        use_threads_for_queue = (worker_type == 'thread')

        image_results = load_and_run_detector_batch(
            model_file=detector_model,
            image_file_names=image_files,
            checkpoint_path=None,
            confidence_threshold=detection_confidence_threshold,
            checkpoint_frequency=-1,
            results=None,
            n_cores=0,
            use_image_queue=True,
            quiet=True,
            image_size=None,
            batch_size=detector_batch_size,
            include_image_size=False,
            include_image_timestamp=False,
            include_exif_tags=None,
            loader_workers=detector_worker_threads,
            preprocess_on_image_queue=True,
            use_threads_for_queue=use_threads_for_queue
        )

        # Write image results to temporary file
        image_output_file = detector_output_file.replace('.json', '_images.json')
        write_results_to_file(image_results,
                              image_output_file,
                              relative_path_base=source_folder,
                              detector_file=detector_model)

        print('Image detection results written to {}'.format(image_output_file))
        files_to_merge.append(image_output_file)

    # ...if we had images to process

    # Process videos if necessary
    if len(video_files) > 0:

        print('Running MegaDetector on {} videos...'.format(len(video_files)))

        # Set up video processing options
        video_options = ProcessVideoOptions()
        video_options.model_file = detector_model
        video_options.input_video_file = source_folder
        video_options.output_json_file = detector_output_file.replace('.json', '_videos.json')
        video_options.json_confidence_threshold = detection_confidence_threshold
        video_options.frame_sample = frame_sample
        video_options.time_sample = time_sample
        video_options.recursive = True

        # Process videos
        process_videos(video_options)

        print('Video detection results written to {}'.format(video_options.output_json_file))
        files_to_merge.append(video_options.output_json_file)

    # ...if we had videos to process

    # Merge results if we have both images and videos
    if len(files_to_merge) > 1:
        print('Merging image and video detection results...')
        combine_batch_output_files(files_to_merge, detector_output_file)
        print('Merged detection results written to {}'.format(detector_output_file))
    elif len(files_to_merge) == 1:
        # Just rename the single file
        if files_to_merge[0] != detector_output_file:
            if os.path.isfile(detector_output_file):
                print('Detector file {} exists, over-writing'.format(detector_output_file))
                os.remove(detector_output_file)
            os.rename(files_to_merge[0], detector_output_file)
        print('Detection results written to {}'.format(detector_output_file))

# ...def _run_detection_step(...)


def _run_classification_step(detector_results_file: str,
                             merged_results_file: str,
                             source_folder: str,
                             classifier_model: str = DEFAULT_CLASSIFIER_MODEL,
                             classifier_batch_size: int = DEFAULT_CLASSIFIER_BATCH_SIZE,
                             classifier_worker_threads: int = DEFAULT_LOADER_WORKERS,
                             detection_confidence_threshold: float = \
                                DEFAULT_DETECTION_CONFIDENCE_THRESHOLD_FOR_CLASSIFICATION,
                             enable_rollup: bool = True,
                             country: str = None,
                             admin1_region: str = None,
                             top_n_scores: int = DEFAULT_TOP_N_SCORES,
                             worker_type: str = DEFAULT_WORKER_TYPE,
                             include_raw_classifications: bool = False,
                             rollup_target_confidence: float = DEFAULT_ROLLUP_TARGET_CONFIDENCE):
    """
    Run SpeciesNet classification on detections from MegaDetector results.

    Args:
        detector_results_file (str): path to MegaDetector output .json file
        merged_results_file (str): path to which we should write the merged results
        source_folder (str): source folder for resolving relative paths
        classifier_model (str, optional): classifier model identifier
        classifier_batch_size (int, optional): batch size for classification
        classifier_worker_threads (int, optional): number of worker threads
        detection_confidence_threshold (float, optional): classify detections above this threshold
        enable_rollup (bool, optional): whether to apply taxonomic rollup
        country (str, optional): country code for geofencing
        admin1_region (str, optional): admin1 region (typically a state code) for geofencing
        top_n_scores (int, optional): maximum number of scores to include for each detection
        worker_type (str, optional): type of worker parallelization ("thread" or "process")
        include_raw_classifications (bool, optional): include raw (pre-rollup/geofence)
            classification scores in output
        rollup_target_confidence (float, optional): target confidence threshold for taxonomic
            rollup.  Ignored if enable_rollup is False.
    """

    print('Starting classification step...')

    # Load MegaDetector results
    print('Reading detection results from {}'.format(detector_results_file))

    with open(detector_results_file, 'r') as f:
        detector_results = json.load(f)

    print('Classification step loaded detection results for {} images'.format(
        len(detector_results['images'])))

    images = detector_results['images']
    if len(images) == 0:
        raise ValueError('No images found in detector results')

    print('Using SpeciesNet classifier: {}'.format(classifier_model))

    # Set multiprocessing start method to 'spawn' for CUDA compatibility
    if worker_type == 'process':
        original_start_method = multiprocessing.get_start_method()
        if original_start_method != 'spawn':
            multiprocessing.set_start_method('spawn', force=True)
            print('Set multiprocessing start method to spawn (was {})'.format(
                original_start_method))

    ## Set up multiprocessing queues

    # This queue receives lists of image filenames (and associated detection results)
    # from the "main" thread (the one you're reading right now).  Items are pulled off
    # of this queue by producer workers (on _crop_producer_func), where the corresponding
    # images are loaded from disk and preprocessed into crops.
    image_queue = JoinableQueue(maxsize= \
                                classifier_worker_threads * MAX_IMAGE_QUEUE_SIZE_PER_WORKER)

    # This queue receives cropped images from producers (on _crop_producer_func); those
    # crops are pulled off of this queue by the consumer (on _crop_consumer_func).
    batch_queue = Queue(maxsize=MAX_BATCH_QUEUE_SIZE)

    # This is not really used as a queue, rather it's just used to send all the results
    # at once from the consumer process to the main process (the one you're reading right
    # now).
    results_queue = Queue()

    WorkerClass = Thread if worker_type == 'thread' else Process # noqa

    # When using threads, pre-load classifiers in the main thread to avoid PyTorch FX issues
    # with loading models in worker threads. When using processes, pass None and let each
    # process load its own classifier.
    if worker_type == 'thread':
        # Producer classifier (CPU only, used for preprocessing)
        producer_classifier = SpeciesNetClassifier(classifier_model, device='cpu')
        # Consumer classifier (GPU if available, used for inference)
        consumer_classifier = SpeciesNetClassifier(classifier_model)
    else:
        producer_classifier = None
        consumer_classifier = None

    # Start producer workers
    producers = []
    for i_worker in range(classifier_worker_threads):
        p = WorkerClass(target=_crop_producer_func,
                        args=(image_queue, batch_queue, classifier_model,
                              detection_confidence_threshold, source_folder, i_worker,
                              producer_classifier))
        p.start()
        producers.append(p)


    ## Start consumer worker

    consumer = WorkerClass(target=_crop_consumer_func,
                           args=(batch_queue, results_queue, classifier_model,
                                 classifier_batch_size, classifier_worker_threads,
                                 enable_rollup, country, admin1_region, consumer_classifier,
                                 rollup_target_confidence))
    consumer.start()

    # This will block every time the queue reaches its maximum depth, so for
    # very small jobs, this will not be a useful progress bar.
    with tqdm(total=len(images),desc='Classification') as pbar:
        for image_data in images:
            image_queue.put(image_data)
            pbar.update()

    # Send sentinel signals to producers
    for _ in range(classifier_worker_threads):
        image_queue.put(None)

    # Wait for all work to complete
    image_queue.join()

    print('Finished waiting for input queue')


    ## Wait for results

    classification_results = results_queue.get()


    ## Clean up processes

    for p in producers:
        p.join()
    consumer.join()

    print('Finished waiting for workers')


    ## Format results and write output

    class CategoryState:
        """
        Helper class to manage classification category IDs.
        """

        def __init__(self):

            self.next_category_id = 0

            # Maps common name to string-int IDs
            self.common_name_to_id = {}

            # Maps string-ints to common names, as per format standard
            self.classification_categories = {}

            # Maps string-ints to latin taxonomy strings, as per format standard
            self.classification_category_descriptions = {}

        def _get_category_id(self, class_name):
            """
            Get an integer-valued category ID for the 7-token string [class_name],
            creating a new one if necessary.
            """

            # E.g.:
            #
            # "cb553c4e-42c9-4fe0-9bd0-da2d6ed5bfa1;mammalia;carnivora;canidae;urocyon;littoralis;island fox"
            tokens = class_name.split(';')
            assert len(tokens) == 7
            taxonomy_string = ';'.join(tokens[1:6])
            common_name = tokens[6]
            if len(common_name) == 0:
                common_name = taxonomy_string

            if common_name not in self.common_name_to_id:
                self.common_name_to_id[common_name] = str(self.next_category_id)
                self.classification_categories[str(self.next_category_id)] = common_name
                # Store the full seven-token string, rather than the shortened five-token string, for
                # compatibility with what is expected by the classification_postprocessing module.
                # self.classification_category_descriptions[str(self.next_category_id)] = taxonomy_string
                self.classification_category_descriptions[str(self.next_category_id)] = class_name
                self.next_category_id += 1

            category_id = self.common_name_to_id[common_name]

            return category_id

    # ...class CategoryState

    category_state = CategoryState()

    # Merge classification results back into detector results with proper category IDs
    for image_data in images:

        image_file = image_data['file']

        if ('detections' not in image_data) or (image_data['detections'] is None):
            continue

        detections = image_data['detections']

        if image_file not in classification_results:
            continue

        image_classifications = classification_results[image_file]

        for detection_index, detection in enumerate(detections):

            if detection_index in image_classifications:

                result = image_classifications[detection_index]

                if 'failure' in result:
                    # Add failure to the image, not the detection
                    if 'failure' not in image_data:
                        image_data['failure'] = result['failure']
                    else:
                        image_data['failure'] += ';' + result['failure']
                else:

                    # Convert class names to category IDs
                    classification_pairs = []
                    raw_classification_pairs = []

                    scores = [x[1] for x in result['classifications']]
                    assert is_list_sorted(scores, reverse=True)

                    # Only report the requested number of scores per detection
                    if len(result['classifications']) > top_n_scores:
                        result['classifications'] = \
                            result['classifications'][0:top_n_scores]

                    if len(result['raw_classifications']) > top_n_scores:
                        result['raw_classifications'] = \
                            result['raw_classifications'][0:top_n_scores]

                    for class_name, score in result['classifications']:

                        category_id = category_state._get_category_id(class_name)
                        score = round_float(score, precision=CONF_DIGITS)
                        classification_pairs.append([category_id, score])

                    for class_name, score in result['raw_classifications']:

                        category_id = category_state._get_category_id(class_name)
                        score = round_float(score, precision=CONF_DIGITS)
                        raw_classification_pairs.append([category_id, score])

                    # Add classifications to the detection
                    detection['classifications'] = classification_pairs
                    if include_raw_classifications:
                        detection['raw_classifications'] = raw_classification_pairs

                # ...if this classification contains a failure

            # ...if this detection has classification information

        # ...for each detection

    # ...for each image

    # Update metadata in the results
    if 'info' not in detector_results:
        detector_results['info'] = {}

    detector_results['info']['classifier'] = classifier_model
    detector_results['info']['classification_completion_time'] = time.strftime(
        '%Y-%m-%d %H:%M:%S')

    # Add classification category mapping
    detector_results['classification_categories'] = \
        category_state.classification_categories
    detector_results['classification_category_descriptions'] = \
        category_state.classification_category_descriptions

    print('Writing output file')

    # Write results
    write_json(merged_results_file, detector_results)

    if verbose:
        print('Classification results written to {}'.format(merged_results_file))

# ...def _run_classification_step(...)


#%% Main function

def run_md_and_speciesnet(options):
    """
    Main entry point, runs MegaDetector and SpeciesNet on a folder.  See
    RunMDSpeciesNetOptions for available arguments.

    Args:
        options (RunMDSpeciesNetOptions): options controlling MD and SN inference
    """

    # Set global verbose flag
    global verbose
    verbose = options.verbose

    # Also set the run_detector_batch verbose flag
    run_detector_batch.verbose = verbose

    # Validate arguments
    if not os.path.isdir(options.source):
        raise ValueError(
            'Source folder does not exist: {}'.format(options.source))

    if (options.admin1_region is not None) and (options.country is None):
        raise ValueError('--admin1_region requires --country to be specified')

    if options.skip_images and options.skip_video:
        raise ValueError('Cannot skip both images and videos')

    if (options.frame_sample is not None) and (options.time_sample is not None):
        raise ValueError('--frame_sample and --time_sample are mutually exclusive')
    if (options.frame_sample is None) and (options.time_sample is None):
        options.time_sample = DEFAULT_SECONDS_PER_VIDEO_FRAME

    if options.worker_type not in ('thread','process'):
        raise ValueError('Unknown worker type {}'.format(options.worker_type))

    # Set up intermediate file folder
    if options.intermediate_file_folder:
        temp_folder = options.intermediate_file_folder
        os.makedirs(temp_folder, exist_ok=True)
    else:
        temp_folder = make_temp_folder(subfolder='run_md_and_speciesnet')

    start_time = time.time()

    print('Processing folder: {}'.format(options.source))
    print('Output file: {}'.format(options.output_file))
    print('Intermediate file folder: {}'.format(temp_folder))

    assert options.overwrite_handling in ('overwrite','error','skip'), \
        'Unknown overwrite_handling value {}'.format(options.overwrite_handling)

    if os.path.isdir(options.output_file):
        raise ValueError('Output file {} exists, but is a directory'.format(
            options.output_file))
    if os.path.isfile(options.output_file):
        if options.overwrite_handling == 'overwrite':
            print('Over-writing existing output file {}'.format(options.output_file))
        elif options.overwrite_handling == 'error':
            raise ValueError('Output file {} exists, and overwrite_handling is "error"'.format(
                options.output_file))
        elif options.ovwrite_handling == 'skip':
            print('Bypassing proecssing: output file {} exists, and overwrite_handling is "skip"'.format(
                options.output_file))
            return

    # Verify that we can create the output file
    test_file_write(options.output_file)

    # Determine detector output file path
    if options.detections_file is not None:
        detector_output_file = options.detections_file
        if VALIDATE_DETECTION_FILE:
            print('Using existing detections file: {}'.format(detector_output_file))
            validation_options = ValidateBatchResultsOptions()
            validation_options.check_image_existence = True
            validation_options.relative_path_base = options.source
            validation_options.raise_errors = True
            validate_batch_results(detector_output_file,options=validation_options)
            print('Validated detections file')
        else:
            print('Bypassing validation of {}'.format(options.detections_file))
    else:
        detector_output_file = os.path.join(temp_folder, 'detector_output.json')

        # Run MegaDetector
        _run_detection_step(
            source_folder=options.source,
            detector_output_file=detector_output_file,
            detector_model=options.detector_model,
            detector_batch_size=options.detector_batch_size,
            detection_confidence_threshold=options.detection_confidence_threshold_for_output,
            detector_worker_threads=options.loader_workers,
            skip_images=options.skip_images,
            skip_video=options.skip_video,
            frame_sample=options.frame_sample,
            time_sample=options.time_sample,
            worker_type=options.worker_type
        )

    # Run SpeciesNet
    _run_classification_step(
        detector_results_file=detector_output_file,
        merged_results_file=options.output_file,
        source_folder=options.source,
        classifier_model=options.classification_model,
        classifier_batch_size=options.classifier_batch_size,
        classifier_worker_threads=options.loader_workers,
        detection_confidence_threshold=options.detection_confidence_threshold_for_classification,
        enable_rollup=(not options.norollup),
        country=options.country,
        admin1_region=options.admin1_region,
        worker_type=options.worker_type,
        include_raw_classifications=options.include_raw_classifications,
        rollup_target_confidence=options.rollup_target_confidence
    )

    elapsed_time = time.time() - start_time
    print(
        'Processing complete in {}'.format(humanfriendly.format_timespan(elapsed_time)))
    print('Results written to: {}'.format(options.output_file))

    # Clean up intermediate files if requested
    if (not options.keep_intermediate_files) and \
       (not options.intermediate_file_folder) and \
       (not options.detections_file):
        try:
            os.remove(detector_output_file)
        except Exception as e:
            print('Warning: error removing temporary output file {}: {}'.format(
                detector_output_file, str(e)))

# ...def run_md_and_speciesnet(...)


#%% Command-line driver

def main():
    """
    Command-line driver for run_md_and_speciesnet.py
    """

    if 'speciesnet' not in sys.modules:
        print('It looks like the speciesnet package is not available, try "pip install speciesnet"')
        if not is_sphinx_build():
            sys.exit(-1)

    parser = argparse.ArgumentParser(
        description='Run MegaDetector and SpeciesNet on a folder of images/videos',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    # Required arguments
    parser.add_argument('source',
                        help='Folder containing images and/or videos to process')
    parser.add_argument('output_file',
                        help='Output file for results (JSON format)')

    # Optional arguments
    parser.add_argument('--detector_model',
                        default=DEFAULT_DETECTOR_MODEL,
                        help='MegaDetector model identifier')
    parser.add_argument('--classification_model',
                        default=DEFAULT_CLASSIFIER_MODEL,
                        help='SpeciesNet classifier model identifier')
    parser.add_argument('--detector_batch_size',
                        type=int,
                        default=DEFAULT_DETECTOR_BATCH_SIZE,
                        help='Batch size for MegaDetector inference')
    parser.add_argument('--classifier_batch_size',
                        type=int,
                        default=DEFAULT_CLASSIFIER_BATCH_SIZE,
                        help='Batch size for SpeciesNet classification')
    parser.add_argument('--loader_workers',
                        type=int,
                        default=DEFAULT_LOADER_WORKERS,
                        help='Number of worker threads for preprocessing')
    parser.add_argument('--detection_confidence_threshold_for_classification',
                        type=float,
                        default=DEFAULT_DETECTION_CONFIDENCE_THRESHOLD_FOR_CLASSIFICATION,
                        help='Classify detections above this threshold')
    parser.add_argument('--detection_confidence_threshold_for_output',
                        type=float,
                        default=DEFAULT_DETECTION_CONFIDENCE_THRESHOLD_FOR_OUTPUT,
                        help='Include detections above this threshold in the output')
    parser.add_argument('--intermediate_file_folder',
                        default=None,
                        help='Folder for intermediate files (default: system temp)')
    parser.add_argument('--keep_intermediate_files',
                        action='store_true',
                        help='Keep intermediate files (e.g. detection-only results file)')
    parser.add_argument('--norollup',
                        action='store_true',
                        help='Disable taxonomic rollup')
    parser.add_argument('--rollup_target_confidence',
                        type=float,
                        default=DEFAULT_ROLLUP_TARGET_CONFIDENCE,
                        help='Target confidence threshold for taxonomic rollup ' + \
                             f'(default {DEFAULT_ROLLUP_TARGET_CONFIDENCE}), only ' + \
                             'used when geofencing is disabled')
    parser.add_argument('--country',
                        default=None,
                        help='Country code (ISO 3166-1 alpha-3) for geofencing')
    parser.add_argument('--admin1_region', '--state',
                        default=None,
                        help='Admin1 region/state code for geofencing')
    parser.add_argument('--detections_file',
                        default=None,
                        help='Path to existing MegaDetector output file (skips detection step)')
    parser.add_argument('--skip_video',
                        action='store_true',
                        help='Ignore videos, only process images')
    parser.add_argument('--skip_images',
                        action='store_true',
                        help='Ignore images, only process videos')
    parser.add_argument('--frame_sample',
                        type=int,
                        default=None,
                        help='Sample every Nth frame from videos (mutually exclusive with --time_sample)')
    parser.add_argument('--time_sample',
                        type=float,
                        default=None,
                        help='Sample frames every N seconds from videos (default {})'.\
                            format(DEFAULT_SECONDS_PER_VIDEO_FRAME) + \
                            ' (mutually exclusive with --frame_sample)')
    parser.add_argument('--verbose',
                        action='store_true',
                        help='Enable additional debug output')
    parser.add_argument('--include_raw_classifications',
                        action='store_true',
                        help='Include raw (pre-rollup/geofence) classification scores in output')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    options = RunMDSpeciesNetOptions()
    args_to_object(args,options)

    run_md_and_speciesnet(options)

# ...def main(...)


if __name__ == '__main__':
    main()



"""

rename_images.py

Copies images from a possibly-nested folder structure to a flat folder structure, including EXIF
timestamps in each filename.  Loosely equivalent to camtrapR's imageRename() function.

"""

#%% Imports and constants

import os
import sys
import argparse

from megadetector.utils.path_utils import \
    find_images, insert_before_extension, parallel_copy_files
from megadetector.data_management.read_exif import \
    ReadExifOptions, read_exif_from_folder


#%% Functions

def rename_images(input_folder,
                  output_folder,
                  dry_run=False,
                  verbose=False,
                  read_exif_options=None,
                  n_copy_workers=8):
    """
    Copies images from a possibly-nested folder structure to a flat folder structure,
    including EXIF timestamps in each filename.  Loosely equivalent to camtrapR's
    imageRename() function.

    Args:
        input_folder (str): the folder to search for images, always recursive
        output_folder (str): the folder to which we will copy images; cannot be the
            same as [input_folder]
        dry_run (bool, optional): only map images, don't actually copy
        verbose (bool, optional): enable additional debug output
        read_exif_options (ReadExifOptions, optional): parameters controlling the reading of
            EXIF information
        n_copy_workers (int, optional): number of parallel threads to use for copying

    Returns:
        dict: a dict mapping relative filenames in the input folder to relative filenames in the output
        folder
    """

    assert os.path.isdir(input_folder), 'Input folder {} does not exist'.format(
        input_folder)

    if not dry_run:
        os.makedirs(output_folder,exist_ok=True)

    # Read exif information
    if read_exif_options is None:
        read_exif_options = ReadExifOptions()

    read_exif_options.tags_to_include = ['DateTime','Model',
                                         'Make','ExifImageWidth',
                                         'ExifImageHeight','DateTimeOriginal']
    read_exif_options.verbose = False

    exif_info = read_exif_from_folder(input_folder=input_folder,
                                      output_file=None,
                                      options=read_exif_options,
                                      filenames=None,recursive=True)

    print('Read EXIF information for {} images'.format(len(exif_info)))

    filename_to_exif_info = {info['file_name']:info for info in exif_info}

    image_files = find_images(input_folder,return_relative_paths=True,convert_slashes=True,recursive=True)

    for fn in image_files:
        assert fn in filename_to_exif_info, 'No EXIF info available for {}'.format(fn)

    input_fn_relative_to_output_fn_relative = {}

    # fn_relative = image_files[0]
    for fn_relative in image_files:

        input_fn_abs = os.path.join(input_folder,fn_relative)
        image_exif_info = filename_to_exif_info[fn_relative]
        if 'exif_tags' in image_exif_info:
            image_exif_info = image_exif_info['exif_tags']

        if image_exif_info is None or \
            'DateTimeOriginal' not in image_exif_info or \
            image_exif_info['DateTimeOriginal'] is None:

            dt_tag = 'unknown_datetime'
            print('Warning: no datetime for {}'.format(fn_relative))

        else:

            dt_tag = str(image_exif_info['DateTimeOriginal']).replace(':','-').replace(' ','_').strip()

        flat_filename = fn_relative.replace('\\','/').replace('/','_')

        output_fn_relative = insert_before_extension(flat_filename,dt_tag)

        input_fn_relative_to_output_fn_relative[fn_relative] = output_fn_relative

    if not dry_run:

        input_fn_abs_to_output_fn_abs = {}

        for input_fn_relative in input_fn_relative_to_output_fn_relative:

            output_fn_relative = input_fn_relative_to_output_fn_relative[input_fn_relative]
            input_fn_abs = os.path.join(input_folder,input_fn_relative)
            output_fn_abs = os.path.join(output_folder,output_fn_relative)
            input_fn_abs_to_output_fn_abs[input_fn_abs] = output_fn_abs

        parallel_copy_files(input_file_to_output_file=input_fn_abs_to_output_fn_abs,
                            max_workers=n_copy_workers,
                            use_threads=True,
                            overwrite=True,
                            verbose=verbose)

    # ...if this is not a dry run

    return input_fn_relative_to_output_fn_relative

# ...def rename_images()


#%% Interactive driver

if False:

    pass

    #%% Configure options

    input_folder = r'G:\camera_traps\camera_trap_images\2018.05.04'
    output_folder = r'g:\temp\rename-test-out'
    dry_run = False
    verbose = True
    read_exif_options = ReadExifOptions()
    read_exif_options.tags_to_include = ['DateTime','Model','Make',
                                         'ExifImageWidth','ExifImageHeight',
                                         'DateTimeOriginal']
    read_exif_options.n_workers = 8
    read_exif_options.verbose = verbose
    n_copy_workers = 8


    #%% Programmatic execution

    input_fn_relative_to_output_fn_relative = rename_images(input_folder,
                                                            output_folder,
                                                            dry_run=dry_run,
                                                            verbose=verbose,
                                                            read_exif_options=read_exif_options,
                                                            n_copy_workers=n_copy_workers)


#%% Command-line driver

def main(): # noqa

    parser = argparse.ArgumentParser(
        description='Copies images from a possibly-nested folder structure to a flat folder structure, ' + \
            'adding datetime information from EXIF to each filename')

    parser.add_argument(
        'input_folder',
        type=str,
        help='The folder to search for images, always recursive')

    parser.add_argument(
        'output_folder',
        type=str,
        help='The folder to which we should write the flattened image structure')

    parser.add_argument(
        '--dry_run',
        action='store_true',
        help="Only map images, don't actually copy")

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    rename_images(args.input_folder,args.output_folder,dry_run=args.dry_run,
                  verbose=True,read_exif_options=None)

if __name__ == '__main__':
    main()



"""

run_inference_with_yolov5_val.py

Runs a folder of images through MegaDetector (or another YOLOv5/YOLOv8 model) with YOLO's
val.py, converting the output to the standard MD format.  The reasons this script exists,
as an alternative to the standard run_detector_batch.py are:

* This script provides access to YOLO's test-time augmentation tools.
* This script serves a reference implementation: by any reasonable definition, YOLOv5's
  val.py produces the "correct" result for any image, since it matches what was used in
  training.
* This script works for any Ultralytics detection model, including YOLOv8 models

YOLOv5's val.py uses each file's base name as a unique identifier, which doesn't work
when you have typical camera trap images like:

* a/b/c/RECONYX0001.JPG
* d/e/f/RECONYX0001.JPG

...both of which would just be "RECONYX0001.JPG".  So this script jumps through a bunch of
hoops to put a symlinks in a flat folder, run YOLOv5 on that folder, and map the results back
to the real files.

If you are running a YOLOv5 model, this script currently requires the caller to supply the path
where a working YOLOv5 install lives, and assumes that the current conda environment is all set up for
YOLOv5.  If you are running a YOLOv8 model, the folder doesn't matter, but it assumes that ultralytics
tools are available in the current environment.

By default, this script uses symlinks to format the input images in a way that YOLO's
val.py likes, as per above.  This requires admin privileges on Windows... actually technically this
only requires permissions to create symbolic links, but I've never seen a case where someone has
that permission and *doesn't* have admin privileges.  If you are running this script on
Windows and you don't have admin privileges, use --no_use_symlinks, which will make copies of images,
rather than using symlinks.

"""

#%% Imports

import os
import sys
import uuid
import glob
import tempfile
import shutil
import json
import copy
import argparse

from tqdm import tqdm

from megadetector.utils import path_utils
from megadetector.utils import process_utils
from megadetector.utils import string_utils
from megadetector.utils.ct_utils import args_to_object

from megadetector.utils.ct_utils import is_iterable, split_list_into_fixed_size_chunks
from megadetector.utils import ct_utils
from megadetector.utils.path_utils import path_is_abs
from megadetector.data_management import yolo_output_to_md_output
from megadetector.detection.run_detector import try_download_known_detector
from megadetector.postprocessing.combine_batch_outputs import combine_batch_output_files

default_image_size_with_augmentation = int(1280 * 1.3)
default_image_size_with_no_augmentation = 1280


#%% Options class

class YoloInferenceOptions:
    """
    Parameters that control the behavior of run_inference_with_yolov5_val(), including
    the input/output filenames.
    """

    def __init__(self):

        ## Required-ish ##

        #: Folder of images to process (can be None if image_filename_list contains absolute paths)
        self.input_folder = None

        #: If this is None, [input_folder] can't be None, we'll process all images in [input_folder].
        #:
        #: If this is not None, and [input_folder] is not None, this should be a list of relative image
        #: paths within [input_folder] to process, or a .txt or .json file containing a list of
        #: relative image paths.
        #:
        #: If this is not None, and [input_folder] is None, this should be a list of absolute image
        #: paths, or a .txt or .json file containing a list of absolute image paths.
        self.image_filename_list = None

        #: Model filename (ending in .pt), or a well-known model name (e.g. "MDV5A")
        self.model_filename = None

        #: .json output file, in MD results format
        self.output_file = None


        ## Optional ##

        #: Required for older YOLOv5 inference, not for newer ulytralytics/YOLOv8 inference
        self.yolo_working_folder = None

        #: Currently 'yolov5' and 'ultralytics' are supported, and really these are proxies for
        #: "the yolov5 repo" and "the ultralytics repo".
        self.model_type = 'yolov5'

        #: Image size to use; this is a single int, which in ultralytics's terminology means
        #: "scale the long side of the image to this size, and preserve aspect ratio".
        #:
        #: If None, will choose based on whether augmentation is enabled.
        self.image_size = None

        #: Detections below this threshold will not be included in the output file
        self.conf_thres = '0.001'

        #: Batch size... has no impact on results, but may create memory issues if you set
        #: this to large values
        self.batch_size = 1

        #: Device string: typically '0' for GPU 0, '1' for GPU 1, etc., or 'cpu'
        self.device_string = '0'

        #: Should we enable test-time augmentation?
        self.augment = False

        #: Should we enable half-precision inference?
        self.half_precision_enabled = None

        #: Where should we stash the temporary symlinks (or copies) used to give unique identifiers to image
        # files?
        #:
        #: If this is None, we'll create a folder in system temp space.
        self.symlink_folder = None

        #: Should we use symlinks to give unique identifiers to image files (vs. copies)?
        self.use_symlinks = True

        #: How should we guarantee that YOLO IDs (base filenames) are unique?  Choices are:
        #:
        #: * 'verify': assume image IDs are unique, but verify and error if they're not
        #: * 'links': create symlinks (or copies, depending on use_symlinks) to enforce uniqueness
        #: * 'auto': check whether IDs are unique, create links if necessary
        self.unique_id_strategy = 'links'

        #: Temporary folder to stash intermediate YOLO results.
        #:
        #: If this is None, we'll create a folder in system temp space.
        self.yolo_results_folder = None

        #: Should we remove the symlink folder when we're done?
        self.remove_symlink_folder = True

        #: Should we remove the intermediate results folder when we're done?
        self.remove_yolo_results_folder = True

        #: These are deliberately offset from the standard MD categories; YOLOv5
        #: needs categories IDs to start at 0.
        #:
        #: This can also be a string that points to any class mapping file supported
        #: by read_classes_from_yolo_dataset_file(): a YOLO dataset.yaml file, a text
        #: file with a list of classes, or a .json file with an ID --> name dict
        self.yolo_category_id_to_name = {0:'animal',1:'person',2:'vehicle'}

        #: What should we do if the output file already exists?
        #:
        #: Can be 'error', 'skip', or 'overwrite'.
        self.overwrite_handling = 'skip'

        #: If True, we'll do a dry run that lets you preview the YOLO val command, without
        #: actually running it.
        self.preview_yolo_command_only = False

        #: By default, if any errors occur while we're copying images or creating symlinks, it's
        #: game over.  If this is True, those errors become warnings, and we plow ahead.
        self.treat_copy_failures_as_warnings = False

        #: Save YOLO console output
        self.save_yolo_debug_output = False

        #: Whether to search for images recursively within [input_folder]
        #:
        #: Ignored if a list of files is provided.
        self.recursive = True

        #: Maximum number of images to run in a single chunk
        self.checkpoint_frequency = None

        #: By default, if we're creating symlinks to images, we append a unique job ID to the
        #: symlink folder.  If the caller is 100% sure that the symlink folder can be re-used
        #: across calls, this can be set to False.
        self.append_job_id_to_symlink_folder = True

        #: By default, we turn category ID 0 coming out of the YOLO .json file
        #: into category 1 in the MD-formatted .json file.
        self.offset_yolo_category_ids = True

    # ...def __init__()

# ...YoloInferenceOptions()


#%% Support functions

def _clean_up_temporary_folders(options,
                                symlink_folder,yolo_results_folder,
                                symlink_folder_is_temp_folder,yolo_folder_is_temp_folder):
    """
    Remove temporary symlink/results folders, unless the caller requested that we leave them in place.
    """

    if options.remove_symlink_folder:
        shutil.rmtree(symlink_folder)
    elif symlink_folder_is_temp_folder:
        print('Warning: using temporary symlink folder {}, but not removing it'.format(
            symlink_folder))

    if options.remove_yolo_results_folder:
        shutil.rmtree(yolo_results_folder)
    elif yolo_folder_is_temp_folder:
        print('Warning: using temporary YOLO results folder {}, but not removing it'.format(
            yolo_results_folder))


def get_stats_for_category(filename,category='all'):
    """
    Retrieve statistics for a category from the YOLO console output
    stored in [filenam].

    Args:
        filename (str): a text file containing console output from a YOLO val run
        category (str, optional): a category name

    Returns:
        dict: a dict with fields n_images, n_labels, P, R, mAP50, and mAP50-95
    """

    with open(filename,'r',encoding='utf-8') as f:
        lines = f.readlines()

    # This is just a hedge to make sure there isn't some YOLO version floating
    # around that used different IoU thresholds in the console output.
    found_map50 = False
    found_map5095 = False

    for line in lines:

        s = line.strip()

        if ' map50 ' in s.lower() or ' map@.5 ' in s.lower():
            found_map50 = True
        if 'map50-95' in s.lower() or 'map@.5:.95' in s.lower():
            found_map5095 = True

        if not s.startswith(category):
            continue

        tokens = s.split(' ')
        tokens_filtered = list(filter(None,tokens))

        if len(tokens_filtered) != 7:
            continue

        assert found_map50 and found_map5095, \
            'Parsing error in YOLO console output file {}'.format(filename)

        to_return = {}
        to_return['category'] = category
        assert category == tokens_filtered[0]
        to_return['n_images'] = int(tokens_filtered[1])
        to_return['n_labels'] = int(tokens_filtered[2])
        to_return['P'] = float(tokens_filtered[3])
        to_return['R'] = float(tokens_filtered[4])
        to_return['mAP50'] = float(tokens_filtered[5])
        to_return['mAP50-95'] = float(tokens_filtered[6])
        return to_return

    # ...for each line

    return None


#%% Main function

def run_inference_with_yolo_val(options):
    """
    Runs a folder of images through MegaDetector (or another YOLOv5/YOLOv8 model) with YOLO's
    val.py, converting the output to the standard MD format.

    Args:
        options (YoloInferenceOptions): all the parameters used to control this process,
            including filenames; see YoloInferenceOptions for details
    """

    ##%% Input and path handling

    default_options = YoloInferenceOptions()

    for k in options.__dict__.keys():
        if k not in default_options.__dict__:
            # Print warnings about unexpected variables, except for things like
            # "no_append_job_id_to_symlink_folder", which just negate existing objects
            if not k.startswith('no_'):
                print('Warning: unexpected variable {} in options object'.format(k))

    if options.model_type == 'yolov8':

        print('Warning: model type "yolov8" supplied, "ultralytics" is the preferred model ' + \
              'type string for YOLOv8 models')
        options.model_type = 'ultralytics'

    if (options.model_type == 'yolov5') and ('yolov8' in options.model_filename.lower()):
        print('\n\n*** Warning: model type set as "yolov5", but your model filename contains "yolov8"... ' + \
              'did you mean to use --model_type yolov8?" ***\n\n')

    if options.yolo_working_folder is None:
        assert options.model_type == 'ultralytics', \
            'A working folder is required to run YOLOv5 val.py'
    else:
        assert os.path.isdir(options.yolo_working_folder), \
            'Could not find working folder {}'.format(options.yolo_working_folder)

    if options.half_precision_enabled is not None:
        assert options.half_precision_enabled in (0,1), \
            'Invalid value {} for --half_precision_enabled (should be 0 or 1)'.format(
                options.half_precision_enabled)

    # If the model filename is a known model string (e.g. "MDv5A", download the model if necessary)
    model_filename = try_download_known_detector(options.model_filename)

    assert os.path.isfile(model_filename), \
        'Could not find model file {}'.format(model_filename)

    assert (options.input_folder is not None) or (options.image_filename_list is not None), \
        'You must specify a folder and/or a file list'

    if options.input_folder is not None:
        assert os.path.isdir(options.input_folder), 'Could not find input folder {}'.format(
            options.input_folder)

    if os.path.exists(options.output_file):
        if options.overwrite_handling == 'skip':
            print('Warning: output file {} exists, skipping'.format(options.output_file))
            return
        elif options.overwrite_handling == 'overwrite':
            print('Warning: output file {} exists, overwriting'.format(options.output_file))
        elif options.overwrite_handling == 'error':
            raise ValueError('Output file {} exists'.format(options.output_file))
        else:
            raise ValueError('Unknown output handling method {}'.format(options.overwrite_handling))

    output_dir = os.path.dirname(options.output_file)
    if len(output_dir) > 0:
        os.makedirs(output_dir, exist_ok=True)

    if options.input_folder is not None:
        options.input_folder = options.input_folder.replace('\\','/')


    ##%% Other input handling

    if isinstance(options.yolo_category_id_to_name,str):

        assert os.path.isfile(options.yolo_category_id_to_name)
        yolo_dataset_file = options.yolo_category_id_to_name
        options.yolo_category_id_to_name = \
            yolo_output_to_md_output.read_classes_from_yolo_dataset_file(yolo_dataset_file)
        print('Loaded {} category mappings from {}'.format(
            len(options.yolo_category_id_to_name),yolo_dataset_file))

    temporary_folder = None
    symlink_folder_is_temp_folder = False
    yolo_folder_is_temp_folder = False

    job_id = str(uuid.uuid1())

    def get_job_temporary_folder(tf):
        if tf is not None:
            return tf
        tempdir_base = tempfile.gettempdir()
        tf = os.path.join(tempdir_base,'md_to_yolo','md_to_yolo_' + job_id)
        os.makedirs(tf,exist_ok=True)
        return tf

    symlink_folder = options.symlink_folder
    yolo_results_folder = options.yolo_results_folder

    if symlink_folder is None:
        temporary_folder = get_job_temporary_folder(temporary_folder)
        symlink_folder = os.path.join(temporary_folder,'symlinks')
        symlink_folder_is_temp_folder = True

    if yolo_results_folder is None:
        temporary_folder = get_job_temporary_folder(temporary_folder)
        yolo_results_folder = os.path.join(temporary_folder,'yolo_results')
        yolo_folder_is_temp_folder = True

    if options.append_job_id_to_symlink_folder:
        # Attach a GUID to the symlink folder, regardless of whether we created it
        symlink_folder_inner = os.path.join(symlink_folder,job_id)
    else:
        print('Re-using existing symlink folder {}'.format(symlink_folder))
        symlink_folder_inner = symlink_folder

    os.makedirs(symlink_folder_inner,exist_ok=True)
    os.makedirs(yolo_results_folder,exist_ok=True)


    ##%% Enumerate images

    image_files_relative = None
    image_files_absolute = None

    # If the caller just provided a folder, not a list of files...
    if options.image_filename_list is None:

        assert options.input_folder is not None and os.path.isdir(options.input_folder), \
            'Could not find input folder {}'.format(options.input_folder)
        image_files_relative = path_utils.find_images(options.input_folder,
                                                      recursive=options.recursive,
                                                      return_relative_paths=True,
                                                      convert_slashes=True)
        image_files_absolute = [os.path.join(options.input_folder,fn) for \
                                fn in image_files_relative]

    else:

        # If the caller provided a list of image files (rather than a filename pointing
        # to a list of image files)...
        if is_iterable(options.image_filename_list) and not isinstance(options.image_filename_list,str):

            image_files_relative = options.image_filename_list

        # If the caller provided a filename pointing to a list of image files...
        else:

            assert isinstance(options.image_filename_list,str), \
                'Unrecognized image filename list object type: {}'.format(options.image_filename_list)
            assert os.path.isfile(options.image_filename_list), \
                'Could not find image filename list file: {}'.format(options.image_filename_list)
            ext = os.path.splitext(options.image_filename_list)[-1].lower()
            assert ext in ('.json','.txt'), \
                'Unrecognized image filename list file extension: {}'.format(options.image_filename_list)
            if ext == '.json':
                with open(options.image_filename_list,'r') as f:
                    image_files_relative = json.load(f)
                    assert is_iterable(image_files_relative)
            else:
                assert ext == '.txt'
                with open(options.image_filename_list,'r') as f:
                    image_files_relative = f.readlines()
                    image_files_relative = [s.strip() for s in image_files_relative]

        # ...whether the image filename list was supplied as list vs. a filename

        if options.input_folder is None:

            image_files_absolute = image_files_relative

        else:

            # The list should be relative filenames
            for fn in image_files_relative:
                assert not path_is_abs(fn), \
                    'When providing a folder and a list, paths in the list should be relative'

            image_files_absolute = \
                [os.path.join(options.input_folder,fn) for fn in image_files_relative]

        for fn in image_files_absolute:
            assert os.path.isfile(fn), 'Could not find image file {}'.format(fn)

    # ...whether the caller supplied a list of filenames

    image_files_absolute = [fn.replace('\\','/') for fn in image_files_absolute]

    del image_files_relative


    ##%% Recurse if necessary to handle checkpoints

    if options.checkpoint_frequency is not None and options.checkpoint_frequency > 0:

        chunks = split_list_into_fixed_size_chunks(image_files_absolute,options.checkpoint_frequency)

        chunk_output_files = []

        # i_chunk = 0; chunk_files_abs = chunks[i_chunk]
        for i_chunk,chunk_files_abs in enumerate(chunks):

            print('Processing {} images from chunk {} of {}'.format(
                len(chunk_files_abs),i_chunk,len(chunks)))

            chunk_options = copy.deepcopy(options)

            # Run each chunk without checkpointing
            chunk_options.checkpoint_frequency = None

            if options.input_folder is not None:
                chunk_files_relative = \
                    [os.path.relpath(fn,options.input_folder) for fn in chunk_files_abs]
                chunk_options.image_filename_list = chunk_files_relative
            else:
                chunk_options.image_filename_list = chunk_files_abs

            chunk_options.image_filename_list = \
                [fn.replace('\\','/') for fn in chunk_options.image_filename_list]

            chunk_string = 'chunk_{}'.format(str(i_chunk).zfill(5))
            chunk_options.yolo_results_folder = yolo_results_folder + '_' + chunk_string
            chunk_options.symlink_folder = symlink_folder + '_' + chunk_string

            # Put the output file in the parent job's scratch folder
            chunk_output_file = os.path.join(yolo_results_folder,chunk_string + '_results_md_format.json')
            chunk_output_files.append(chunk_output_file)
            chunk_options.output_file = chunk_output_file

            if os.path.isfile(chunk_output_file):

                print('Chunk output file {} exists, checking completeness'.format(chunk_output_file))

                with open(chunk_output_file,'r') as f:
                    chunk_results = json.load(f)
                images_in_this_chunk_results_file = [im['file'] for im in chunk_results['images']]
                assert len(images_in_this_chunk_results_file) == len(chunk_options.image_filename_list), \
                    f'Expected {len(chunk_options.image_filename_list)} images in ' + \
                    f'chunk results file {chunk_output_file}, found {len(images_in_this_chunk_results_file)}, ' + \
                     'possibly this is left over from a previous job?'
                for fn in images_in_this_chunk_results_file:
                    assert fn in chunk_options.image_filename_list, \
                        f'Unexpected image {fn} in chunk results file {chunk_output_file}, ' + \
                         'possibly this is left over from a previous job?'

                print('Chunk output file {} exists and is complete, skipping this chunk'.format(
                    chunk_output_file))

            # ...if the outptut file exists

            else:

                run_inference_with_yolo_val(chunk_options)

            # ...if we do/don't have to run this chunk

            assert os.path.isfile(chunk_options.output_file)

        # ...for each chunk

        # Merge
        _ = combine_batch_output_files(input_files=chunk_output_files,
                                 output_file=options.output_file,
                                 require_uniqueness=True,
                                 verbose=True)

        # Validate
        with open(options.output_file,'r') as f:
            combined_results = json.load(f)
        assert len(combined_results['images']) == len(image_files_absolute), \
            'Expected {} images in merged output file, found {}'.format(
                len(image_files_absolute),len(combined_results['images']))

        # Clean up
        _clean_up_temporary_folders(options,
                                    symlink_folder,yolo_results_folder,
                                    symlink_folder_is_temp_folder,yolo_folder_is_temp_folder)

        return

    # ...if we need to make recursive calls for file chunks


    ##%% Create symlinks (or copy images) to give a unique ID to each image

    # Maps YOLO image IDs (base filename without extension as it will appear in YOLO .json output)
    # to the *original full path* for each image (not the symlink path).
    image_id_to_file = {}

    # Maps YOLO image IDs (base filename without extension as it will appear in YOLO .json output)
    # to errors, including errors that happen before we run the model at all (e.g. file access errors).
    image_id_to_error = {}

    create_links = True

    if options.unique_id_strategy == 'links':

        create_links = True

    else:

        assert options.unique_id_strategy in ('auto','verify'), \
            'Unknown unique ID strategy {}'.format(options.unique_id_strategy)

        image_ids_are_unique = True

        for i_image,image_fn in tqdm(enumerate(image_files_absolute),total=len(image_files_absolute)):

            image_id = os.path.splitext(os.path.basename(image_fn))[0]

            # Is this image ID unique?
            if image_id in image_id_to_file:
                if options.unique_id_strategy == 'verify':
                    raise ValueError('"verify" specified for image uniqueness, but ' +
                                     'image ID {} occurs more than once:\n\n{}\n\n{}'.format(
                                         image_id,image_fn,image_id_to_file[image_id]))
                else:
                    assert options.unique_id_strategy == 'auto'
                    image_ids_are_unique = False
                    image_id_to_file = {}
                    break

            image_id_to_file[image_id] = image_fn

        # ...for each image

        if image_ids_are_unique:

            print('"{}" specified for image uniqueness and images are unique, skipping links'.format(
                options.unique_id_strategy))
            assert len(image_id_to_file) == len(image_files_absolute)
            create_links = False

        else:

            assert options.unique_id_strategy == 'auto'
            create_links = True
            link_type = 'copies'
            if options.use_symlinks:
                link_type = 'links'
            print('"auto" specified for image uniqueness and images are not unique, defaulting to {}'.format(
                link_type))

    # ...which unique ID strategy?

    if create_links:

        if options.use_symlinks:
            print('Creating {} symlinks in {}'.format(len(image_files_absolute),symlink_folder_inner))
        else:
            print('Symlinks disabled, copying {} images to {}'.format(len(image_files_absolute),symlink_folder_inner))

        link_full_paths = []

        # i_image = 0; image_fn = image_files_absolute[i_image]
        for i_image,image_fn in tqdm(enumerate(image_files_absolute),total=len(image_files_absolute)):

            ext = os.path.splitext(image_fn)[1]
            image_fn_without_extension = os.path.splitext(image_fn)[0]

            # YOLO .json output identifies images by the base filename without the extension
            image_id = str(i_image).zfill(10)
            image_id_to_file[image_id] = image_fn
            symlink_name = image_id + ext
            symlink_full_path = os.path.join(symlink_folder_inner,symlink_name)
            link_full_paths.append(symlink_full_path)

            # If annotation files exist, link those too; only useful if we're reading the computed
            # mAP value, but it doesn't hurt.
            annotation_fn = image_fn_without_extension + '.txt'
            annotation_file_exists = False
            if os.path.isfile(annotation_fn):
                annotation_file_exists = True
                annotation_symlink_name = image_id + '.txt'
                annotation_symlink_full_path = os.path.join(symlink_folder_inner,annotation_symlink_name)

            try:

                if options.use_symlinks:
                    path_utils.safe_create_link(image_fn,symlink_full_path)
                    if annotation_file_exists:
                        path_utils.safe_create_link(annotation_fn,annotation_symlink_full_path)
                else:
                    shutil.copyfile(image_fn,symlink_full_path)
                    if annotation_file_exists:
                        shutil.copyfile(annotation_fn,annotation_symlink_full_path)

            except Exception as e:

                error_string = str(e)
                image_id_to_error[image_id] = error_string

                # Always break if the user is trying to create symlinks on Windows without
                # permission, 100% of images will always fail in this case.
                if ('a required privilege is not held by the client' in error_string.lower()) or \
                   (not options.treat_copy_failures_as_warnings):

                       print('\nError copying/creating link for input file {}: {}'.format(
                           image_fn,error_string))

                       raise

                else:

                    print('Warning: error copying/creating link for input file {}: {}'.format(
                        image_fn,error_string))
                    continue

            # ...except

        # ...for each image

    # ...if we need to create links/copies


    ##%% Create the dataset file if necessary

    # This may have been passed in as a string, but at this point, we should have
    # loaded the dataset file.
    assert isinstance(options.yolo_category_id_to_name,dict)

    # Category IDs need to be continuous integers starting at 0
    category_ids = sorted(list(options.yolo_category_id_to_name.keys()))
    assert category_ids[0] == 0
    assert len(category_ids) == 1 + category_ids[-1]

    yolo_dataset_file = os.path.join(yolo_results_folder,'dataset.yaml')
    yolo_image_list_file = os.path.join(yolo_results_folder,'images.txt')

    with open(yolo_image_list_file,'w') as f:

        if create_links:
            image_files_to_write = link_full_paths
        else:
            image_files_to_write = image_files_absolute

        for fn_abs in image_files_to_write:
            # At least in YOLOv5 val (need to verify for YOLOv8 val), filenames in this
            # text file are treated as relative to the text file itself if they start with
            # "./", otherwise they're treated as absolute paths.  Since we don't want to put this
            # text file in the image folder, we'll use absolute paths.
            # fn_relative = os.path.relpath(fn_abs,options.input_folder)
            # f.write(fn_relative + '\n')
            f.write(fn_abs + '\n')

    if create_links:
        inference_folder = symlink_folder_inner
    else:
        # This doesn't matter, but it has to be a valid path
        inference_folder = options.yolo_results_folder

    with open(yolo_dataset_file,'w') as f:

        f.write('path: {}\n'.format(inference_folder))
        # These need to be valid paths, even if you're not using them, and "." is always safe
        f.write('train: .\n')
        f.write('val: .\n')
        f.write('test: {}\n'.format(yolo_image_list_file))
        f.write('\n')
        f.write('nc: {}\n'.format(len(options.yolo_category_id_to_name)))
        f.write('\n')
        f.write('names:\n')
        for category_id in category_ids:
            assert isinstance(category_id,int)
            f.write('  {}: {}\n'.format(category_id,
                                        options.yolo_category_id_to_name[category_id]))


    ##%% Prepare Python command or YOLO CLI command

    if options.image_size is None:
        if options.augment:
            image_size = default_image_size_with_augmentation
        else:
            image_size = default_image_size_with_no_augmentation
    else:
        image_size = options.image_size

    image_size_string = str(round(image_size))

    if options.model_type == 'yolov5':

        cmd = 'python val.py --task test --data "{}"'.format(yolo_dataset_file)
        cmd += ' --weights "{}"'.format(model_filename)
        cmd += ' --batch-size {} --imgsz {} --conf-thres {}'.format(
            options.batch_size,image_size_string,options.conf_thres)
        cmd += ' --device "{}" --save-json'.format(options.device_string)
        cmd += ' --project "{}" --name "{}" --exist-ok'.format(yolo_results_folder,'yolo_results')

        # This is the NMS IoU threshold
        # cmd += ' --iou-thres 0.6'

        if options.augment:
            cmd += ' --augment'

        # --half is a store_true argument for YOLOv5's val.py
        if (options.half_precision_enabled is not None) and (options.half_precision_enabled == 1):
            cmd += ' --half'

        # Sometimes useful for debugging
        # cmd += ' --save_conf --save_txt'

    elif options.model_type == 'ultralytics':

        if options.augment:
            augment_string = 'augment'
        else:
            augment_string = ''

        cmd = 'yolo val {} model="{}" imgsz={} batch={} data="{}" project="{}" name="{}" device="{}"'.\
            format(augment_string,model_filename,image_size_string,options.batch_size,
                   yolo_dataset_file,yolo_results_folder,'yolo_results',options.device_string)
        cmd += ' save_json exist_ok'

        if (options.half_precision_enabled is not None):
            if options.half_precision_enabled == 1:
                cmd += ' --half=True'
            else:
                assert options.half_precision_enabled == 0
                cmd += ' --half=False'

        # Sometimes useful for debugging
        # cmd += ' save_conf save_txt'

    else:

        raise ValueError('Unrecognized model type {}'.format(options.model_type))

    # print(cmd); import clipboard; clipboard.copy(cmd)


    ##%% Run YOLO command

    if options.yolo_working_folder is not None:
        current_dir = os.getcwd()
        os.chdir(options.yolo_working_folder)

    print('Running YOLO inference command:\n{}\n'.format(cmd))

    if options.preview_yolo_command_only:

        if options.remove_symlink_folder:
            try:
                print('Removing YOLO symlink folder {}'.format(symlink_folder))
                shutil.rmtree(symlink_folder)
            except Exception:
                print('Warning: error removing symlink folder {}'.format(symlink_folder))
                pass
        if options.remove_yolo_results_folder:
            try:
                print('Removing YOLO results folder {}'.format(yolo_results_folder))
                shutil.rmtree(yolo_results_folder)
            except Exception:
                print('Warning: error removing YOLO results folder {}'.format(yolo_results_folder))
                pass

        # sys.exit()
        return

    execution_result = process_utils.execute_and_print(cmd,encoding='utf-8',verbose=True)
    assert execution_result['status'] == 0, 'Error running {}'.format(options.model_type)
    yolo_console_output = execution_result['output']

    if options.save_yolo_debug_output:

        with open(os.path.join(yolo_results_folder,'yolo_console_output.txt'),'w',encoding='utf-8') as f:
            for s in yolo_console_output:
                f.write(s + '\n')
        ct_utils.write_json(os.path.join(yolo_results_folder,'image_id_to_file.json'), image_id_to_file)
        ct_utils.write_json(os.path.join(yolo_results_folder,'image_id_to_error.json'), image_id_to_error)


    # YOLO console output contains lots of ANSI escape codes, remove them for easier parsing
    yolo_console_output = [string_utils.remove_ansi_codes(s) for s in yolo_console_output]

    # Find errors that occurred during the initial corruption check; these will not be included in the
    # output.  Errors that occur during inference will be handled separately.
    yolo_read_failures = []

    for line in yolo_console_output:

        #
        # Lines indicating read failures look like:
        #
        # For ultralytics val:
        #
        # line = "val: WARNING ⚠️ /a/b/c/d.jpg: ignoring corrupt image/label: [Errno 13] Permission denied: '/a/b/c/d.jpg'"
        #
        # For yolov5 val.py:
        #
        # line = "test: WARNING: a/b/c/d.jpg: ignoring corrupt image/label: cannot identify image file '/a/b/c/d.jpg'"
        #
        # In both cases, when we are using symlinks, the first filename is the symlink name, the
        # second filename is the target, e.g.:
        #
        # line = "test: WARNING: /tmp/md_to_yolo/md_to_yolo_xyz/symlinks/xyz/0000000004.jpg: ignoring corrupt image/label: cannot identify image file '/tmp/md-tests/md-test-images/corrupt-images/real-file.jpg'"
        #
        # Windows example:
        #
        # line = "test: WARNING: g:\\temp\\md-test-images\\corrupt-images\\irfanview-can-still-read-me-caltech_camera_traps_5a0e37cc-23d2-11e8-a6a3-ec086b02610b.jpg: ignoring corrupt image/label: cannot identify image file 'g:\\\\temp\\\\md-test-images\\\\corrupt-images\\\\irfanview-can-still-read-me-caltech_camera_traps_5a0e37cc-23d2-11e8-a6a3-ec086b02610b.jpg'"
        #

        line = line.replace('⚠️',':')
        if 'ignoring corrupt image/label' in line:

            line_tokens = line.split('ignoring corrupt image/label')
            assert len(line_tokens) == 2

            tokens = line_tokens[0].split(':') # ,maxsplit=3)
            tokens = [s.strip() for s in tokens]

            # ['test', ' WARNING', ' a/b/c/d.jpg', ' ']
            assert len(tokens[-1]) == 0
            tokens = tokens[:-1]
            assert 'warning' in tokens[1].lower()

            if len(tokens) == 3:
                image_name = tokens[2].strip()
            else:
                # Windows filenames have one extra colon
                assert len(tokens) == 4
                assert len(tokens[2]) == 1
                image_name = ':'.join(tokens[2:4])

            yolo_read_failures.append(image_name)

        # ...if this line indicated a corrupt image

    # ...for each line in the console output

    # image_file = yolo_read_failures[0]
    for image_file in yolo_read_failures:
        image_id = os.path.splitext(os.path.basename(image_file))[0]
        assert image_id in image_id_to_file, 'Unexpected image ID {}'.format(image_id)
        if image_id not in image_id_to_error:
            image_id_to_error[image_id] = 'YOLO read failure'

    if options.yolo_working_folder is not None:
        os.chdir(current_dir)


    ##%% Convert results to MD format

    json_files = glob.glob(yolo_results_folder + '/yolo_results/*.json')
    assert len(json_files) == 1
    yolo_json_file = json_files[0]

    # Map YOLO image IDs to paths
    image_id_to_relative_path = {}
    for image_id in image_id_to_file:
        fn = image_id_to_file[image_id].replace('\\','/')
        assert path_is_abs(fn)
        if options.input_folder is not None:
            assert os.path.isdir(options.input_folder)
            assert options.input_folder in fn, 'Internal error: base folder {} not in file {}'.format(
                options.input_folder,fn)
            relative_path = os.path.relpath(fn,options.input_folder)
        else:
            # We'll use the absolute path as a relative path, and pass '/'
            # as the base path in this case.
            relative_path = fn
        image_id_to_relative_path[image_id] = relative_path

    # Are we working with a base folder?
    if options.input_folder is not None:
        assert os.path.isdir(options.input_folder)
        image_base = options.input_folder
    else:
        image_base = '/'

    yolo_output_to_md_output.yolo_json_output_to_md_output(
        yolo_json_file=yolo_json_file,
        image_folder=image_base,
        output_file=options.output_file,
        yolo_category_id_to_name=options.yolo_category_id_to_name,
        detector_name=os.path.basename(model_filename),
        image_id_to_relative_path=image_id_to_relative_path,
        image_id_to_error=image_id_to_error,
        offset_yolo_class_ids=options.offset_yolo_category_ids)


    ##%% Clean up

    _clean_up_temporary_folders(options,
                                symlink_folder,yolo_results_folder,
                                symlink_folder_is_temp_folder,yolo_folder_is_temp_folder)

# ...def run_inference_with_yolo_val()


#%% Command-line driver

def main(): # noqa

    options = YoloInferenceOptions()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        'model_filename',type=str,
        help='model file name')
    parser.add_argument(
        'input_folder',type=str,
        help='folder on which to recursively run the model, or a .json or .txt file ' + \
             'containing a list of absolute image paths')
    parser.add_argument(
        'output_file',type=str,
        help='.json file where output will be written')

    parser.add_argument(
        '--image_filename_list',type=str,default=None,
        help='.json or .txt file containing a list of relative image filenames within [input_folder]')
    parser.add_argument(
        '--yolo_working_folder',type=str,default=None,
        help='folder in which to execute val.py (not necessary for YOLOv8 inference)')
    parser.add_argument(
        '--image_size', default=None, type=int,
        help='image size for model execution (default {} when augmentation is enabled, else {})'.format(
            default_image_size_with_augmentation,default_image_size_with_no_augmentation))
    parser.add_argument(
        '--conf_thres', default=options.conf_thres, type=float,
        help='confidence threshold for including detections in the output file (default {})'.format(
            options.conf_thres))
    parser.add_argument(
        '--batch_size', default=options.batch_size, type=int,
        help='inference batch size (default {})'.format(options.batch_size))
    parser.add_argument(
        '--half_precision_enabled', default=None, type=int,
        help='use half-precision-inference (1 or 0) (default is the underlying model\'s default, ' + \
             'probably full for YOLOv8 and half for YOLOv5')
    parser.add_argument(
        '--device_string', default=options.device_string, type=str,
        help='CUDA device specifier, typically "0" or "1" for CUDA devices, "mps" for ' + \
             'M1/M2 devices, or "cpu" (default {})'.format(
            options.device_string))
    parser.add_argument(
        '--overwrite_handling', default=options.overwrite_handling, type=str,
        help='action to take if the output file exists (skip, error, overwrite) (default {})'.format(
            options.overwrite_handling))
    parser.add_argument(
        '--yolo_dataset_file', default=None, type=str,
        help='YOLOv5 dataset.yaml file from which we should load category information ' + \
            '(otherwise defaults to MD categories)')
    parser.add_argument(
        '--model_type', default=options.model_type, type=str,
        help='model type ("yolov5" or "ultralytics" ("yolov8" behaves the same as "ultralytics")) (default {})'.format(
            options.model_type))

    parser.add_argument('--unique_id_strategy', default=options.unique_id_strategy, type=str,
        help='how should we ensure that unique filenames are passed to the YOLO val script, ' + \
             'can be "verify", "auto", or "links", see options class docs for details (default {})'.format(
                 options.unique_id_strategy))
    parser.add_argument(
        '--symlink_folder', default=None, type=str,
        help='temporary folder for symlinks (defaults to a folder in the system temp dir)')
    parser.add_argument(
        '--yolo_results_folder', default=None, type=str,
        help='temporary folder for YOLO intermediate output (defaults to a folder in the system temp dir)')
    parser.add_argument(
        '--no_use_symlinks', action='store_true',
        help='copy files instead of creating symlinks when preparing the yolo input folder')
    parser.add_argument(
        '--no_remove_symlink_folder', action='store_true',
        help='don\'t remove the temporary folder full of symlinks')
    parser.add_argument(
        '--no_remove_yolo_results_folder', action='store_true',
        help='don\'t remove the temporary folder full of YOLO intermediate files')
    parser.add_argument(
        '--save_yolo_debug_output', action='store_true',
        help='write yolo console output to a text file in the results folder, along with additional debug files')
    parser.add_argument(
        '--checkpoint_frequency', default=options.checkpoint_frequency, type=int,
        help='break the job into chunks with no more than this many images (default {})'.format(
            options.checkpoint_frequency))
    parser.add_argument(
        '--no_append_job_id_to_symlink_folder', action='store_true',
        help="don't append a unique job ID to the symlink folder name")
    parser.add_argument(
        '--nonrecursive', action='store_true',
        help='disable recursive folder processing')
    parser.add_argument(
        '--no_offset_class_ids', action='store_true',
        help='disable class ID offsetting')

    parser.add_argument(
        '--preview_yolo_command_only', action='store_true',
        help='don\'t run inference, just preview the YOLO inference command (still creates symlinks)')

    if options.augment:
        default_augment_enabled = 1
    else:
        default_augment_enabled = 0

    parser.add_argument(
        '--augment_enabled', default=default_augment_enabled, type=int,
        help='enable/disable augmentation (default {})'.format(default_augment_enabled))

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    # If the caller hasn't specified an image size, choose one based on whether augmentation
    # is enabled.
    if args.image_size is None:
        assert args.augment_enabled in (0,1), \
            'Illegal augment_enabled value {}'.format(args.augment_enabled)
        if args.augment_enabled == 1:
            args.image_size = default_image_size_with_augmentation
        else:
            args.image_size = default_image_size_with_no_augmentation
        augment_enabled_string = 'enabled'
        if not args.augment_enabled:
            augment_enabled_string = 'disabled'
        print('Augmentation is {}, using default image size {}'.format(
            augment_enabled_string,args.image_size))

    args_to_object(args, options)

    if args.yolo_dataset_file is not None:
        options.yolo_category_id_to_name = args.yolo_dataset_file

    # The function convention is that input_folder should be None when we want to use a list of
    # absolute paths, but the CLI convention is that the required argument is always valid, whether
    # it's a folder or a list of absolute paths.
    if os.path.isfile(options.input_folder):
        assert options.image_filename_list is None, \
            'image_filename_list should not be specified when input_folder is a file'
        options.image_filename_list = options.input_folder
        options.input_folder = None

    options.recursive = (not options.nonrecursive)
    options.append_job_id_to_symlink_folder = (not options.no_append_job_id_to_symlink_folder)
    options.remove_symlink_folder = (not options.no_remove_symlink_folder)
    options.remove_yolo_results_folder = (not options.no_remove_yolo_results_folder)
    options.use_symlinks = (not options.no_use_symlinks)
    options.augment = (options.augment_enabled > 0)
    options.offset_yolo_category_ids = (not options.no_offset_class_ids)

    del options.nonrecursive
    del options.no_remove_symlink_folder
    del options.no_remove_yolo_results_folder
    del options.no_use_symlinks
    del options.augment_enabled
    del options.yolo_dataset_file
    del options.no_offset_class_ids

    print(options.__dict__)

    run_inference_with_yolo_val(options)

if __name__ == '__main__':
    main()


#%% Interactive driver

if False:


    #%% Debugging

    input_folder = r'g:\temp\md-test-images'
    model_filename = 'MDV5A'
    output_folder = r'g:\temp\yolo-test-out'
    yolo_working_folder = r'c:\git\yolov5-md'
    dataset_file = r"g:\temp\md-test-images\dataset.yaml"
    job_name = 'yolo-debug'
    symlink_folder = os.path.join(output_folder,'symlinks')
    yolo_results_folder = os.path.join(output_folder,'yolo_results')
    model_name = os.path.splitext(os.path.basename(model_filename))[0]

    output_file = os.path.join(output_folder,'{}_{}-md_format.json'.format(
        job_name,model_name))

    options = YoloInferenceOptions()

    options.yolo_working_folder = yolo_working_folder
    options.input_folder = input_folder
    options.output_file = output_file

    options.yolo_category_id_to_name = dataset_file
    options.augment = False
    options.conf_thres = '0.001'
    options.batch_size = 1
    options.device_string = '0'
    options.unique_id_strategy = 'auto'
    options.overwrite_handling = 'overwrite'

    if options.augment:
        options.image_size = round(1280 * 1.3)
    else:
        options.image_size = 1280

    options.model_filename = model_filename

    options.yolo_results_folder = yolo_results_folder # os.path.join(output_folder + 'yolo_results')
    options.symlink_folder = symlink_folder # os.path.join(output_folder,'symlinks')
    options.use_symlinks = False

    options.remove_symlink_folder = True
    options.remove_yolo_results_folder = True

    options.checkpoint_frequency = None

    cmd = f'python run_inference_with_yolov5_val.py {model_filename} {input_folder} ' + \
          f'{output_file} --yolo_working_folder {yolo_working_folder} ' + \
          f' --image_size {options.image_size} --conf_thres {options.conf_thres} ' + \
          f' --batch_size {options.batch_size} ' + \
          f' --symlink_folder {options.symlink_folder} --yolo_results_folder {options.yolo_results_folder} ' + \
          f' --yolo_dataset_file {options.yolo_category_id_to_name} ' + \
          f' --unique_id_strategy {options.unique_id_strategy} --overwrite_handling {options.overwrite_handling}'

    if not options.remove_symlink_folder:
        cmd += ' --no_remove_symlink_folder'
    if not options.remove_yolo_results_folder:
        cmd += ' --no_remove_yolo_results_folder'
    if options.checkpoint_frequency is not None:
        cmd += f' --checkpoint_frequency {options.checkpoint_frequency}'
    if not options.use_symlinks:
        cmd += ' --no_use_symlinks'
    if not options.augment:
        cmd += ' --augment_enabled 0'

    print(cmd)
    execute_in_python = False
    if execute_in_python:
        run_inference_with_yolo_val(options)
    else:
        import clipboard; clipboard.copy(cmd)



    #%% Run inference on a folder

    input_folder = r'g:\temp\tegu-val-mini'.replace('\\','/')
    model_filename = r'g:\temp\usgs-tegus-yolov5x-231003-b8-img1280-e3002-best.pt'
    output_folder = r'g:\temp\tegu-scratch'
    yolo_working_folder = r'c:\git\yolov5-tegus'
    dataset_file = r'g:\temp\dataset.yaml'

    # This only impacts the output file name, it's not passed to the inference function
    job_name = 'yolo-inference-test'

    model_name = os.path.splitext(os.path.basename(model_filename))[0]

    symlink_folder = os.path.join(output_folder,'symlinks')
    yolo_results_folder = os.path.join(output_folder,'yolo_results')

    output_file = os.path.join(output_folder,'{}_{}-md_format.json'.format(
        job_name,model_name))

    options = YoloInferenceOptions()

    options.yolo_working_folder = yolo_working_folder
    options.input_folder = input_folder
    options.output_file = output_file

    pass_image_filename_list = False
    pass_relative_paths = True

    if pass_image_filename_list:
        if pass_relative_paths:
            options.image_filename_list =  [
                r"val#american_cardinal#american_cardinal#CaCa#31W.01_C83#2017-2019#C90 and C83_31W.01#(05) 18AUG17 - 05SEP17 FTC AEG#MFDC1949_000065.JPG",
                r"val#american_cardinal#american_cardinal#CaCa#31W.01_C83#2017-2019#C90 and C83_31W.01#(04) 27JUL17 - 18AUG17 FTC AEG#MFDC1902_000064.JPG"
            ]
        else:
            options.image_filename_list =  [
                r"g:/temp/tegu-val-mini/val#american_cardinal#american_cardinal#CaCa#31W.01_C83#2017-2019#C90 and C83_31W.01#(05) 18AUG17 - 05SEP17 FTC AEG#MFDC1949_000065.JPG",
                r"g:/temp/tegu-val-mini/val#american_cardinal#american_cardinal#CaCa#31W.01_C83#2017-2019#C90 and C83_31W.01#(04) 27JUL17 - 18AUG17 FTC AEG#MFDC1902_000064.JPG"
            ]
    else:
        options.image_filename_list = None

    options.yolo_category_id_to_name = dataset_file
    options.augment = False
    options.conf_thres = '0.001'
    options.batch_size = 1
    options.device_string = '0'
    options.unique_id_strategy = 'auto'
    options.overwrite_handling = 'overwrite'

    if options.augment:
        options.image_size = round(1280 * 1.3)
    else:
        options.image_size = 1280

    options.model_filename = model_filename

    options.yolo_results_folder = yolo_results_folder # os.path.join(output_folder + 'yolo_results')
    options.symlink_folder = symlink_folder # os.path.join(output_folder,'symlinks')
    options.use_symlinks = False

    options.remove_symlink_folder = True
    options.remove_yolo_results_folder = True

    options.checkpoint_frequency = 5

    cmd = f'python run_inference_with_yolov5_val.py {model_filename} {input_folder} ' + \
          f'{output_file} --yolo_working_folder {yolo_working_folder} ' + \
          f' --image_size {options.image_size} --conf_thres {options.conf_thres} ' + \
          f' --batch_size {options.batch_size} ' + \
          f' --symlink_folder {options.symlink_folder} --yolo_results_folder {options.yolo_results_folder} ' + \
          f' --yolo_dataset_file {options.yolo_category_id_to_name} ' + \
          f' --unique_id_strategy {options.unique_id_strategy} --overwrite_handling {options.overwrite_handling}'

    if not options.remove_symlink_folder:
        cmd += ' --no_remove_symlink_folder'
    if not options.remove_yolo_results_folder:
        cmd += ' --no_remove_yolo_results_folder'
    if options.checkpoint_frequency is not None:
        cmd += f' --checkpoint_frequency {options.checkpoint_frequency}'
    if not options.use_symlinks:
        cmd += ' --no_use_symlinks'
    if not options.augment:
        cmd += ' --augment_enabled 0'

    print(cmd)
    execute_in_python = False
    if execute_in_python:
        run_inference_with_yolo_val(options)
    else:
        import clipboard; clipboard.copy(cmd)
